#ifndef Application_h
#define Application_h

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "BaseApplication.h"
#include "Config.h"
#include "CommandLineArguments.h"
#include "CommandLineParser.h"
#include "Counter.h"
#include "IconEngine.h"
#include "TaskBarBase.h"
#include "TaskBarAttributes.h"

//* Main Window singleton object
class Application: public BaseApplication, public Counter
{

    //* Qt meta object declaration
    Q_OBJECT

    public:

    //* constructor
    Application( CommandLineArguments );

    //* create all widgets
    bool realizeWidget( void );

    //* number of bars
    unsigned int numberOfTaskBars( void ) const
    { return taskBars_.size(); }

    //* current taskBar
    TaskBarBase* currentTaskBar( void ) const
    { return _taskBar( _currentAttributes() ); }

    //*@name actions
    //@{

    //* add taskBar
    QAction& addTaskBarAction( void ) const
    { return *addTaskBarAction_; }

    //* remove taskBar
    QAction& removeTaskBarAction( void ) const
    { return *removeTaskBarAction_; }

    //@}

    //*@name application information
    //@{

    //* command line help
    void usage( void ) const;

    //* application name
    virtual QString applicationName( void ) const
    { return "Transparent Taskbar"; }

    //* application icon
    virtual QIcon applicationIcon( void ) const
    { return IconEngine::get( ":/taskbar.png" ); }

    // application version
    virtual QString applicationVersion( void ) const
    { return VERSION; }

    //@}

    //* event filter
    /** needed to monitor closed taskbar, and exit when last one is closed */
    virtual bool eventFilter( QObject*, QEvent* );

    protected Q_SLOTS:

    //* configuration
    virtual void _configuration( void );

    //* add taskBar
    void _addTaskBar( void );

    //* modify task bar attributes
    void _editTaskBar( const TaskBarAttributes& );

    //* remove taskBar
    void _removeTaskBar( bool ask = true );

    //* save taskBars
    void _readTaskBars( void );

    //* save taskBars
    void _writeTaskBars(  bool check = true ) const;

    //* actions called when application is about to quit
    void _aboutToQuit( void );

    //* current taskBar attributes
    void _setCurrentAttributes( const TaskBarAttributes& attributes )
    {
        Debug::Throw( "Application::_setCurrentAttributes.\n" );
        currentAttributes_ = attributes;
    }

    //* process request from application manager
    virtual bool _processCommand( Server::ServerCommand );

    private:

    //* create a new task bar from attributes
    TaskBarBase* _newTaskBar( const TaskBarAttributes& attributes = TaskBarAttributes() );

    //* get taskBar from attributes
    TaskBarBase* _taskBar( const TaskBarAttributes& attributes ) const;

    //* update actions
    void _updateActions( void ) const;

    //* current taskBar attributes
    const TaskBarAttributes& _currentAttributes( void ) const
    { return currentAttributes_; }

    //* current taskBar attributes
    TaskBarAttributes& _currentAttributes( void )
    { return currentAttributes_; }

    //* taskBars
    TaskBarBase::List taskBars_;

    //* TaskBarAttributes for the selected taskBar
    TaskBarAttributes currentAttributes_;

    //*@name actions
    //@{

    //* add taskBar
    QAction* addTaskBarAction_ = nullptr;

    //* remove taskBar
    QAction* removeTaskBarAction_ = nullptr;

    //@}

};

#endif
