/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "TaskBarBase.h"

#include "CompositeEngine.h"
#include "Effects.h"
#include "IconNames.h"
#include "IconEngine.h"
#include "InformationDialog.h"
#include "Menu.h"
#include "QtUtil.h"
#include "QuestionDialog.h"
#include "RoundedRegion.h"
#include "SvgEngine.h"
#include "Singleton.h"
#include "TaskDialog.h"
#include "TaskBarConfiguration.h"
#include "TaskBarDialog.h"
#include "TaskListConfiguration.h"
#include "TaskListDialog.h"
#include "TaskParser.h"
#include "ToolTipWidget.h"
#include "WinUtil.h"
#include "XmlOptions.h"
#include "XcbUtil.h"

#include <QApplication>
#include <QAction>
#include <QDesktopWidget>
#include <QMouseEvent>
#include <QTextStream>

#if HAVE_XCB
#include <X11/Xlib.h>
#include <xcb/xcb.h>
#endif

//_____________________________________
QSize TaskBarBase::emptySize( void ) const
{
    const int extraMargin( 10 );
    QRect rect = fontMetrics().boundingRect( _emptyText() );
    rect.adjust( -extraMargin, -extraMargin, extraMargin, extraMargin );
    rect.adjust( -_margins().left(), -_margins().top(), _margins().right(), _margins().bottom() );
    return rect.size();

}

//_____________________________________
TaskBarBase::TaskBarBase( QWidget *parent ):
    TransparentWidget( parent, Qt::FramelessWindowHint ),
    toolTipWidget_( new ToolTipWidget( this ) ),
    selectedButton_( TaskButton::null() )
{

    Debug::Throw( "TaskBarBase::TaskBarBase.\n" );

    // actions
    _installActions();

    addAction( reloadBackgroundAction_ );

    // reset configuration
    connect( Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(updateConfiguration()) );
    connect( qApp->desktop(), SIGNAL(workAreaResized(int)), SLOT(_desktopResizedDelayed()) );

    toolTipWidget_->setFocusProxy( this );
    setMouseTracking( true );

    //* initialize stored options
    storedOptions_ << Options::Pair( "MIN_BUTTON_SIZE", "" );
    storedOptions_ << Options::Pair( "MAX_BUTTON_SIZE", "" );
    storedOptions_ << Options::Pair( "HIGHLIGHT_ON_ENTER", "" );
    storedOptions_ << Options::Pair( "HIGHLIGHT_ON_CLICK", "" );
    storedOptions_ << Options::Pair( "HIGHLIGHT_INTENSITY", "" );
    storedOptions_ << Options::Pair( "HIGHLIGHT_STEPS", "" );
    storedOptions_ << Options::Pair( "TRANSPARENCY_SHADOW_COLOR", "" );
    storedOptions_ << Options::Pair( "TRANSPARENCY_SHADOW_OFFSET", "" );
    storedOptions_ << Options::Pair( "TRANSPARENCY_FOREGROUND_COLOR", "" );

    connect( &Svg::SvgEngine::get(), SIGNAL(changed()), SLOT(updateConfiguration()) );
    connect( &Svg::SvgEngine::get(), SIGNAL(changed()), SLOT(setBackgroundChanged()) );

}

//____________________________________________
void TaskBarBase::synchronizeAttributes( void )
{

    Debug::Throw( "TaskBarBase::synchronizeAttributes.\n" );

    TaskAttributes::List taskAttributes;
    for( const auto& button:buttons_ )
    {
        // retrieve task attributes
        TaskAttributes attributes( button->attributes() );

        // make sure creation time is valid
        if( !attributes.creation().isValid() )
        {
            if( attributes.modification().isValid() || attributes.lastCalled().isValid() ) attributes.setCreation( qMin(  attributes.modification(), attributes.lastCalled() ) );
            else attributes.setCreation( TimeStamp::now() );
        }

        // store in list
        taskAttributes << attributes;

    }

    // update taskBar attributes
    attributes_.setTaskAttributesList( taskAttributes );

}
//____________________________________________
void TaskBarBase::setAttributes( const TaskBarAttributes& attributes, bool alsoUpdate )
{

    Debug::Throw( "TaskBarBase::setAttributes.\n" );

    // change flags
    for( auto&& iter = flagActions().begin(); iter != flagActions().end(); ++iter )
    { iter.key()->setChecked( attributes.hasFlag( iter.value() ) ); }

    // apply flags
    _toggleStaysOnTop( attributes.hasFlag( TaskBarAttributes::StaysOnTop ) );
    _toggleHideFromTaskBar( attributes.hasFlag( TaskBarAttributes::HideFromTaskBar ) );
    _toggleSticky( attributes.hasFlag( TaskBarAttributes::Sticky ) );
    _toggleRoundCorners( attributes.hasFlag( TaskBarAttributes::RoundCorners ) );

    // AutoHide is enabled only if layout match with position.
    flagAction( TaskBarAttributes::AutoHide )->setEnabled(
        ( (attributes.positionMode()&TaskBarAttributes::PositionTop)    && attributes.layout() ==  TaskBarAttributes::HorizontalTop ) ||
        ( (attributes.positionMode()&TaskBarAttributes::PositionBottom) && attributes.layout() ==  TaskBarAttributes::HorizontalBottom ) ||
        ( (attributes.positionMode()&TaskBarAttributes::PositionLeft)   && attributes.layout() ==  TaskBarAttributes::VerticalLeft ) ||
        ( (attributes.positionMode()&TaskBarAttributes::PositionRight) && attributes.layout() ==  TaskBarAttributes::VerticalRight ) );

    // see if buttons need update
    if( _optionsChanged() || attributes_.needTaskUpdate( attributes ) )
    { _updateTasks( attributes ); }

    // store new attributes
    attributes_ = attributes;

    // tooltip position
    if( attributes_.layout() & (TaskBarAttributes::Horizontal|TaskBarAttributes::Box) ) toolTipWidget_->setPreferredPosition( ToolTipWidget::Bottom );
    else toolTipWidget_->setPreferredPosition( ToolTipWidget::Right );

    if( alsoUpdate )
    {

        // reset offset
        _setOffset( QPoint() );

        // resize from buttons
        _resize();
        _updatePosition();

        // move from attributes
        QtUtil::moveWidget( this, attributes_.position() );
        setBackgroundChanged( true );
        _setNeedUpdate( true );

        // clear tooltip cache
        toolTipWidget_->clearCache();

        update();
    }

}

//___________________________________________________
void TaskBarBase::updateConfiguration( void )
{

    Debug::Throw( "TaskBarBase::updateConfiguration.\n" );

    step_ = XmlOptions::get().get<unsigned int>( "GRID" );
    spacing_ = ( XmlOptions::get().get<unsigned int>( "SPACING" ) );

    // tooltip
    toggleToolTipAction().setChecked( XmlOptions::get().get<bool>("SHOW_TOOLTIPS") );

    highlightOnClickAction().setChecked( XmlOptions::get().get<bool>("HIGHLIGHT_ON_CLICK" ) );
    highlightOnEnterAction().setChecked( XmlOptions::get().get<bool>("HIGHLIGHT_ON_ENTER" ) );
    noHighlightAction().setChecked( !(highlightOnClickAction().isChecked() || highlightOnEnterAction().isChecked() ) );

    // svg update
    _setUseSvg( XmlOptions::get().get<bool>( "USE_SVG" ) );

    if( useSvg_ )
    { Svg::SvgEngine::get().reload(); }

    // margins
    _clearMargins();
    Base::Margins margins;
    if( useSvg_ && Svg::SvgEngine::get().isValid() )
    {
        margins = Svg::SvgEngine::get().margins();
        margins -= Base::Margins( 4 );
    }

    // add option
    margins += Base::Margins( XmlOptions::get().get<int>( "MARGIN" ) );
    _setMargins( margins );

    // outer padding
    _clearOuterPadding();
    if( useSvg_ && Svg::SvgEngine::get().isValid() )
    {

        // read either outer padding or margins from SVG, and adjust
        Base::Margins outerPadding( Svg::SvgEngine::get().outerPadding() );
        if( outerPadding.isNull() )
        {
            outerPadding = Svg::SvgEngine::get().margins();
            if( !outerPadding.isNull() ) outerPadding -= Base::Margins( 4 );
        }

        _setOuterPadding( outerPadding );

    }

    // reload background action
    #if !HAVE_XCB
    reloadBackgroundAction_->setEnabled( useSvg_ );
    #endif
    // reset attributes
    _resetAttributes();

}

//______________________________________________
void TaskBarBase::setMask( QRect mask )
{ setMask( mask, RoundedRegion::All ); }

//______________________________________________
void TaskBarBase::setMask( QRect mask, RoundedRegion::Corners corners )
{
    if(  attributes_.hasFlag( TaskBarAttributes::RoundCorners ) ) Transparency::TransparentWidget::setMask( RoundedRegion( mask, corners ) );
    else {

        // for some reason, setMask does not work unless
        // one remove at least one pixmap from the rect
        // to make it a non rectangular region
        QRegion region( mask );
        region -= QRegion( mask.topLeft().x(), mask.topLeft().y(), 1, 1 );
        Transparency::TransparentWidget::setMask( region );

    }
}

//____________________________________________
void TaskBarBase::_addButton( bool is_separator )
{
    Debug::Throw( "TaskBarBase::_addButton.\n" );
    TaskAttributes attributes;

    if( is_separator ) attributes.setFlag( TaskAttributes::IsSeparator, true );
    else {

        TaskDialog dialog( this );
        if( !dialog.centerOnPointer().exec() ) return;
        attributes = dialog.attributes();
        attributes.setModification();

    }

    // create TaskButton and check validity
    TaskButton::Pointer button( new TaskButton( attributes_, attributes ) );
    if( !button->isValid() )
    {
        InformationDialog( this, tr( "Button is not valid. <Add Button> canceled." ) ).centerOnPointer().exec();
        button->deleteLater();
        return;
    }

    // update task list
    // check selected button
    TaskButton::Pointer selected( selectedButton_ );
    if( !selected ) buttons_.append( button );
    else {

        // try get current button and associated iterator
        auto selectedIter = std::find_if( buttons_.begin(), buttons_.end(), TaskButton::SameAttributesFTor( selected->attributes() ) );

        // advance one, so that the new button is inserted "after" and not before the selected buttons
        if( selectedIter != buttons_.end() ) ++selectedIter;

        // insert button
        buttons_.insert( selectedIter, button );
    }


    // resize and save buttons
    _saveButtons();
    _resetAttributes();

}

//____________________________________________
void TaskBarBase::_removeButton( void )
{
    Debug::Throw( "TaskBarBase::Remove.\n" );

    if( !selectedButton_ )
    {
        InformationDialog( this, tr( "No button selected. <Remove> canceled" ) )
            .setWindowTitle( tr( "Remove Button - Taskbar" ) ).centerOnPointer().exec();
        return;
    }

    // store selected button
    TaskAttributes attributes( selectedButton_->attributes() );

    // confirm
    const QString buffer = attributes.hasFlag( TaskAttributes::IsSeparator ) ?
        tr( "Remove separator ?" ):
        QString( tr( "Remove button '%1' ?" ) ).arg( attributes.name() );

    if( !QuestionDialog( this, buffer ).setWindowTitle( tr( "Remove button - Taskbar" ) ).centerOnPointer().exec() ) return;

    // remove task and save
    TaskButton::List::iterator iter;
    while( ( iter = std::find_if( buttons_.begin(), buttons_.end(), TaskButton::SameAttributesFTor( attributes ) ) ) != buttons_.end() )
    {
        (*iter)->deleteLater();
        buttons_.erase( iter );
    }

    // resize and save
    _saveButtons();
    _resetAttributes();

}

//____________________________________________
void TaskBarBase::_editButton( void )
{

    Debug::Throw( "TaskBarBase::Edit.\n" );
    if( !selectedButton_ )
    {
        InformationDialog( this, tr( "No button selected. <Edit> canceled." ) ).centerOnPointer().exec();
        return;
    }

    // store selectedButton_ in case it changes while filling the dialog
    TaskButton::Pointer button = selectedButton_;

    if( button->attributes().hasFlag( TaskAttributes::IsSeparator ) )
    {
        InformationDialog( this, tr( "Separators cannot be edited." ) ).centerOnPointer().exec();
        return;
    }

    // create dialog
    TaskDialog dialog( this );
    if( !dialog.setAttributes( button->attributes() ).centerOnPointer().exec() ) return;

    // retrieve attributes and update button (if needed)
    TaskAttributes attributes = dialog.attributes();
    if( attributes == button->attributes() ) return;

    // create new button
    attributes.setModification();
    TaskButton new_button( attributes_, attributes );
    if( !new_button.isValid() )
    {
        InformationDialog( this, tr( "New attributes are not valid. <Edit> canceled." ) ).centerOnPointer().exec();
        return;
    }

    // update button
    button->setAttributes( attributes_, attributes );
    _setNeedUpdate(true);
    _saveButtons();
    _resetAttributes();

}

//____________________________________________
void TaskBarBase::_editAttributes( void )
{
    Debug::Throw( "TaskBarBase::_editAttributes.\n" );
    TaskBarDialog dialog( this, attributes_ );
    if( !dialog.exec() ) return;

    TaskBarAttributes attributes( dialog.taskBarConfiguration().attributes() );
    if( attributes == attributes_ )
    { Debug::Throw(0) << "TaskBarBase::_editAttributes - Attributes are unchanged." << endl; }

    emit attributesChanged( attributes );
    return;

}

//____________________________________________
void TaskBarBase::_editTaskList( void )
{
    Debug::Throw( "TaskBarBase::_editTaskList.\n" );
    TaskListDialog dialog( this, attributes_.taskAttributesList() );
    if( !dialog.exec() ) return;

    TaskAttributes::List tasklist( dialog.taskListConfiguration().taskAttributesList() );
    attributes_.setTaskAttributesList( tasklist );
    _resetAttributes();

    emit taskListChanged();

    return;
}

//___________________________________________________
void TaskBarBase::_setEnableMove( bool value )
{
    Debug::Throw( "TaskBarBase::_setEnableMove.\n" );
    if( enableMove_ == value ) return;

    enableMove_ = value;
    if( _moveEnabled() )
    {

        _setCursor( _mouseButton() );
        if( selectedButton_ ) { _leaveButton( selectedButton_ ); }
        _setAction( Minimize );

    } else {

        _setCursor( Qt::NoButton );
        _setAction( Enter );

    }

    update();

}

//________________________________________________
void TaskBarBase::_toggleFlagAction( QAction* action )
{
    Debug::Throw( "TaskBarBase::_toggleFlagAction.\n" );
    ActionMap::const_iterator iter( flagActions().find( action ) );
    Q_ASSERT( iter != flagActions().end() );

    // check if state has changed
    if( action->isChecked() == attributes_.hasFlag( iter.value() ) )
    {
        Debug::Throw() << "TaskBarBase::_toggleFlagAction - flag: " << iter.value() << " unchanged" << endl;
        return;
    }

    // modify attributes accordingly
    TaskBarAttributes attributes( attributes_ );
    attributes.setFlag( iter.value(), action->isChecked() );
    emit attributesChanged( attributes );
    return;

}

//______________________________________________
void TaskBarBase::_toggleHighlight( QAction* action )
{

    Debug::Throw() << "TaskBarBase::_toggleHighlight" << endl;
    if( !action->isChecked() )
    { return; }

    XmlOptions::get().set<bool>( "HIGHLIGHT_ON_CLICK", highlightOnClickAction().isChecked() );
    XmlOptions::get().set<bool>( "HIGHLIGHT_ON_ENTER", highlightOnEnterAction().isChecked() );

    // loop over buttons
    for( const auto& button:buttons_ )
    {

        // skip separators of buttons that already have highlighted pixmaps
        if( button->attributes().hasFlag( TaskAttributes::IsSeparator ) || button->hasHighlightedPixmaps() ) continue;

        // reset button attributes
        // to create/clear highlighted pixmaps
        button->setAttributes( attributes_, button->attributes() );

    }

}

//___________________________________________________
void TaskBarBase::_toggleToolTip( bool state )
{

    Debug::Throw( "TaskBarBase::_toggleToolTip.\n" );
    XmlOptions::get().set<bool>( "SHOW_TOOLTIP", state );

}

//__________________________________________________________
void TaskBarBase::_desktopResizedDelayed( void )
{
    Debug::Throw() << "TransparentWindow::_desktopResizedDelayed" << endl;
    QTimer::singleShot( 1000, this, SLOT(_desktopResized()) );
}

//__________________________________________________________
void TaskBarBase::_desktopResized( void )
{
    Debug::Throw( "TaskBarBase::_destktopResized.\n" );

    if( attributes_.positionMode() == TaskBarAttributes::PositionAny ) return;
    _updatePosition();

    // move from attributes
    QtUtil::moveWidget( this, attributes_.position() );
    setBackgroundChanged( true );
    _setNeedUpdate( true );

    // clear tooltip cache
    toolTipWidget_->clearCache();

    update();

}

//_____________________________________________________
#if HAVE_XCB && QT_VERSION < 0x050000
bool TaskBarBase::x11Event( XEvent* event )
{

    Debug::Throw( "TaskBarBase::x11Event.\n" );

    // cast event to a property event
    XPropertyEvent *propertyEvent = (XPropertyEvent*) event;
    if(
        propertyEvent &&
        propertyEvent->window == winId() &&
        propertyEvent->atom == *XcbUtil::get().atom<xcb_atom_t>( XcbDefines::_NET_WM_STATE ) )
    {

        flagAction( TaskBarAttributes::Sticky )->setChecked(
            ( XcbUtil::get().isSupported( XcbDefines::_NET_WM_STATE_STICKY ) && XcbUtil::get().hasState( this, XcbDefines::_NET_WM_STATE_STICKY ) ) ||
            ( XcbUtil::get().isSupported( XcbDefines::_NET_WM_DESKTOP ) && XcbUtil::get().cardinal( this, XcbDefines::_NET_WM_DESKTOP ) == XcbDefines::ALL_DESKTOPS ) );
        flagAction( TaskBarAttributes::HideFromTaskBar )->setChecked( XcbUtil::get().hasState( this, XcbDefines::_NET_WM_STATE_SKIP_TASKBAR ) );

        return true;

    } else return false;

}
#endif

//______________________________________________
void TaskBarBase::timerEvent( QTimerEvent* event )
{

    Debug::Throw( "TaskBarBase::timerEvent.\n" );

    if( event->timerId() == _updateTimer().timerId() )
    {

        update();
        _updateTimer().stop();

    } else if( event->timerId() == _enableMoveTimer().timerId() ) {

        _enableMoveTimer().stop();
        _setEnableMove();

        #if HAVE_XCB
        if( XcbUtil::get().moveWidget( this, QCursor::pos() ) ) {

            _setMoved( true );
            _setMouseButton( Qt::NoButton );
            attributes_.setPositionMode( TaskBarAttributes::PositionAny );

        }
        #endif

    } else if( event->timerId() == _geometryTimer().timerId() ) {
        _updateGeometry( _action() );
        _geometryTimer().stop();
        update();
    }

    else return TransparentWidget::timerEvent( event );

}

//______________________________________________
void TaskBarBase::resizeEvent( QResizeEvent* event )
{
    if( attributes_.hasFlag( TaskBarAttributes::RoundCorners ) ) _roundCorners();

    if( Transparency::CompositeEngine::get().isAvailable() )
    { _updateBlurRegion( _outerPadding().adjustedRect( rect() ).translated( mapTo( window(), QPoint(0,0) ) ) ); }

    TransparentWidget::resizeEvent( event );

}

//___________________________________________________
void TaskBarBase::enterEvent( QEvent* )
{

    Debug::Throw() << "TaskBarBase::enterEvent - button: " << _mouseButton() << " action: " << _action() << endl;

    // activate
    activateWindow();

    // emit signal
    emit taskBarEntered( attributes_ );

    // reset cursor
    _setCursor( _mouseButton() );

    // try get button under pointer
    _setSelectedButton( _button( mapFromGlobal( QCursor::pos() ) ) );

    // if no button pressed zoom button
    if( _mouseButton() == Qt::NoButton && selectedButton_ && _action() == NoAction )
    {
        _enterButton( selectedButton_ );
        update();

        // show tooltip
        if( toggleToolTipAction().isChecked() &&
            !selectedButton_->attributes().hasFlag( TaskAttributes::IsSeparator ) )
        {
            toolTipWidget_->setTaskAttributes( selectedButton_->attributes() );
            toolTipWidget_->setIndexRect( _toolTipIndexRect( selectedButton_ ).translated( mapToGlobal( QPoint(0,0) ) ) );
            toolTipWidget_->showDelayed();
        }

    }

    // show
    if( !_moveEnabled() ) { _setAction( Enter ); }
    update();

}

//___________________________________________________
void TaskBarBase::leaveEvent( QEvent* )
{

    Debug::Throw( "TaskBarBase::leaveEvent.\n" );

    // reset cursor
    _setCursor( Qt::NoButton );

    // disable move
    if( _mouseButton() == Qt::NoButton ) _setEnableMove( false );

    // unzoom _selectedButton and hide tooltip
    if( selectedButton_ ) _leaveButton( selectedButton_ );
    _setAction( Leave );
    update();

    // stop tooltip timers
    toolTipWidget_->hide();

}

//___________________________________________________
void TaskBarBase::mousePressEvent( QMouseEvent *event )
{

    Debug::Throw( "TaskBarBase::mousePressEvent.\n" );

    // retrieve button
    TaskButton::Pointer current = _button( event->pos() );

    // hide tooltip
    toolTipWidget_->hide();

    // check if button has changed
    if( current && current == selectedButton_ && event->button() == Qt::LeftButton )
    {
        _pressButton( selectedButton_ );
        update();
    }

    // update selection and store button state
    _setSelectedButton( current );
    _setMouseButton( event->button() );
    _setCursor( _mouseButton() );

    // store position if left button (for moving)
    if( _mouseButton() == Qt::LeftButton )
    {

        _setMoved( false );
        if( !( attributes_.hasFlag( TaskBarAttributes::LockPosition ) ) )
        {
            _setClickPosition( event->pos() + QPoint(geometry().topLeft() - frameGeometry().topLeft()) );
            _enableMoveTimer().start( 200, this );
        }

        return;
    }

    // mid button change cursor
    if( _mouseButton() == Qt::MidButton ) {
        return;
    }

    // ask for menu in case of right button
    if( _mouseButton() == Qt::RightButton )
    {

        /*
        when raising the menu, the pointer leaves the window
        and it is likely that we don't get the release event.
        as such we force the button state to Qt::NoButton now
        */
        _setMouseButton( Qt::NoButton );
        return QWidget::mousePressEvent( event );

    }

}

//___________________________________________________
void TaskBarBase::mouseReleaseEvent( QMouseEvent *event )
{
    Debug::Throw( "TaskBarBase::mouseReleaseEvent.\n" );

    // retrieve button
    TaskButton::Pointer current = _button( event->pos() );

    // check if button has changed
    if( current && current == selectedButton_ && event->button() == Qt::LeftButton )
    {
        _releaseButton( selectedButton_ );
        update();
    }

    // left button, run application if not moved; disable move
    if(  _mouseButton() == Qt::LeftButton )
    {

        if( current && !_moved() && !current->attributes().exec() )
        {
            const QString buffer = QString( tr( "Could not execute command:\n%1" ) ).arg( current->attributes().command() );
            InformationDialog( this, buffer ).centerOnPointer().exec();
        }

        _enableMoveTimer().stop();

        // left button disable movement
        if( _moveEnabled() ) _setEnableMove( false );
    }

    // mid button end drag/drop
    if( _mouseButton() == Qt::MidButton )
    { _saveButtons(); }

    // update selection and store button state
    _setCursor( Qt::NoButton );
    _setMoved( false );
    _setSelectedButton( current );
    _setMouseButton( Qt::NoButton );

}

//___________________________________________________
void TaskBarBase::mouseMoveEvent( QMouseEvent *event )
{

    Debug::Throw( "TaskBarBase::mouseMoveEvent.\n" );

    // left button, move the window
    if( _mouseButton() == Qt::LeftButton && !attributes_.hasFlag( TaskBarAttributes::LockPosition ) && _moveEnabled() )
    {
        QPoint true_point(event->globalPos() - _clickPosition() );
        QPoint point(
            _step()*int(true_point.x()/_step()),
            _step()*int(true_point.y()/_step())
            );

        // store position and disable auto_hide
        attributes_.setPosition( point );
        attributes_.setPositionMode( TaskBarAttributes::PositionAny );
        move( point );

        // change move flag
        _setMoved( true );

        return;

    }

    // store mouse position
    _setMousePosition( event->pos() );

    // retrieve button under pointer
    TaskButton::Pointer current = _button( event->pos() );

    // middle button, drag
    if( (!attributes_.hasFlag( TaskBarAttributes::LockButtons )) && _mouseButton() == Qt::MidButton )
    {
        _swap( selectedButton_, current );
        update();
        return;
    }

    // unselect old button
    bool needZoom( false );
    if( selectedButton_ && selectedButton_ != current )
    {
        _leaveButton( selectedButton_ );
        needZoom = true;

        // hide tooltip
        if( _mouseButton() == Qt::NoButton && (!current || current->attributes().hasFlag( TaskAttributes::IsSeparator ) ) )
        { toolTipWidget_->hide(); }

    }

    // tooltip
    if( current && _action() == NoAction && _mouseButton() == Qt::NoButton &&
        toggleToolTipAction().isChecked() &&
        !current->attributes().hasFlag( TaskAttributes::IsSeparator ) )
    {

        if( selectedButton_ != current )
        {
            toolTipWidget_->setTaskAttributes( current->attributes() );
            toolTipWidget_->setIndexRect( _toolTipIndexRect( current ).translated( mapToGlobal( QPoint(0,0) ) ) );
        }

        if( toolTipWidget_->isHidden() ) toolTipWidget_->showDelayed();

    }

    // select new button
    if( current && selectedButton_ != current && _action() == NoAction )
    {
        _setSelectedButton( current );
        _enterButton( selectedButton_ );
        needZoom = true;

    }

    if( needZoom )  update();

}

//________________________________________________
void TaskBarBase::contextMenuEvent( QContextMenuEvent* event )
{

    Debug::Throw( "TaskBarBase::contextMenuEvent.\n" );

    // update actions
    editButtonAction().setEnabled( selectedButton_ && !selectedButton_->attributes().hasFlag( TaskAttributes::IsSeparator ) );
    removeButtonAction().setEnabled( selectedButton_ );
    if( selectedButton_ )
    { removeButtonAction().setText( selectedButton_->attributes().hasFlag( TaskAttributes::IsSeparator ) ? "&Remove separator":"&Remove button" ); }

    _contextMenu().exec( event->globalPos() );

}

//___________________________________________________
void TaskBarBase::_paint( QPaintDevice& device, const QRect& rect )
{

    Debug::Throw( "TaskBarBase::_paint.\n" );

    if( _backgroundChanged() )
    {
        _updateBackgroundPixmap();
        _setNeedUpdate( true );
    }

    // draw either empty taskBar of buttons if any
    if( buttons_.empty() ) { _empty( device, rect ); }
    else if( _action() == Enter ) {

        _drawButtons( device, rect );
        _geometryTimer().start( attributes_.zoomDelay(), this );

    } else if( _drawButtons( device, rect ) ) {

        _updateTimer().start( attributes_.zoomDelay(), this );

    } else if( _action() == Leave || _action() == Minimize ) {

        _drawButtons( device, rect );
        _geometryTimer().start( attributes_.zoomDelay(), this );

    }

}

//________________________________________________
Menu* TaskBarBase::_createContextMenu( void )
{

    Debug::Throw( "TaskBarBase::_createDefaultContextMenu.\n" );
    return new Menu( this );

}

//___________________________________________________
void TaskBarBase::_saveButtons( void )
{

    Debug::Throw( "TaskBarBase::_saveButtons.\n" );
    attributes_.clear();
    for( const auto& button:buttons_ )
    { attributes_.addTaskAttributes( button->attributes() ); }

    emit modified();

    Debug::Throw( "TaskBarBase::_saveButtons - done.\n" );

    return;

}


//___________________________________________________
void TaskBarBase::_clearButtons( void )
{

    Debug::Throw( "TaskBarBase::_clearButtons.\n" );
    for( const auto& button:buttons_ )
    { button->deleteLater(); }

    buttons_.clear();

}

//___________________________________________________
void TaskBarBase::_setDimensions( void )
{
    const TaskBarAttributes::Layout &layout( attributes_.layout() );
    if( layout & TaskBarAttributes::Horizontal )
    {

        nLines_ = 1;
        nColumns_ = buttons_.size();

    } else if( layout & TaskBarAttributes::Vertical ) {

        nLines_ = buttons_.size();
        nColumns_ = 1;

    } else if( layout & TaskBarAttributes::Box ) {

        nColumns_ = attributes_.numColumns();
        nLines_ = buttons_.size()/nColumns_;
        while( nLines_*nColumns_ < buttons_.size() ) nLines_ ++;

    }

}

//_______________________________________________________________________________
void TaskBarBase::_enterButton( TaskButton::Pointer button )
{
    Debug::Throw( "TaskBarBase::_enterButton.\n" );
    if( highlightOnEnterAction().isChecked() || ( _mouseButton() == Qt::LeftButton && highlightOnClickAction().isChecked() ) )
    { button->highlight( true ); }

    button->setDirection( TaskButton::Bigger );
}

//_______________________________________________________________________________
void TaskBarBase::_leaveButton( TaskButton::Pointer button )
{
    Debug::Throw( "TaskBarBase::_leaveButton.\n" );
    button->setDirection( TaskButton::Smaller );
    if( highlightOnEnterAction().isChecked() || highlightOnClickAction().isChecked() ) button->highlight( false );
}

//_______________________________________________________________________________
void TaskBarBase::_pressButton( TaskButton::Pointer button )
{
    Debug::Throw( "TaskBarBase::_pressButton.\n" );
    if( highlightOnClickAction().isChecked() ) {
        button->highlight( true );
        button->setDirection( TaskButton::Bigger );
    } else {
        button->setDirection( TaskButton::Smaller );
    }

}

//_______________________________________________________________________________
void TaskBarBase::_releaseButton( TaskButton::Pointer button )
{ button->setDirection( TaskButton::Bigger ); }

//_______________________________________________________________________________
void TaskBarBase::_swap( TaskButton::Pointer first, TaskButton::Pointer second )
{
    // check buttons
    if( !( first && second && first != second ) ) return;

    const int firstIndex( buttons_.indexOf( first ) );
    const int secondIndex( buttons_.indexOf( second ) );
    buttons_.swap( firstIndex, secondIndex );
}

//________________________________________________
void TaskBarBase::_toggleStaysOnTop( bool state )
{
    Debug::Throw( "TaskBarBase::_toggleStaysOnTop.\n" );

    // check that widget is top level
    if( parentWidget() ) return;

    #if HAVE_XCB

    // change property
    XcbUtil::get().changeState( this, XcbDefines::_NET_WM_STATE_STAYS_ON_TOP, state );
    XcbUtil::get().changeState( this, XcbDefines::_NET_WM_STATE_ABOVE, state );

    #else

    bool visible( !isHidden() );
    if( visible )
    {
        hide();
        qApp->processEvents();
    }

    // Qt implementation
    if( state ) setWindowFlags( windowFlags() | Qt::WindowStaysOnTopHint );
    else setWindowFlags( windowFlags() & ~Qt::WindowStaysOnTopHint );

    if( visible ) show();

    #endif

}

//________________________________________________
void TaskBarBase::_toggleHideFromTaskBar( bool state )
{

    Debug::Throw() << "TaskBarBase::_toggleHideFromTaskBar - state: " << state << endl;

    // check that widget is top level
    if( parentWidget() ) return;

    #if HAVE_XCB
    XcbUtil::get().changeState( this, XcbDefines::_NET_WM_STATE_SKIP_TASKBAR, state );
    XcbUtil::get().changeState( this, XcbDefines::_NET_WM_STATE_SKIP_PAGER, state );
    #endif

    #if defined(Q_OS_WIN)
    WinUtil( this ).toggleHideFromTaskBar( state );
    #endif


    return;

}

//________________________________________________
void TaskBarBase::_toggleSticky( bool state )
{

    Debug::Throw() << "TaskBarBase::_toggleSticky - state: " << state << endl;

    // check that widget is top level
    if( parentWidget() ) return;

    #if HAVE_XCB
    if( XcbUtil::get().isSupported( XcbDefines::_NET_WM_STATE_STICKY ) )
    {

        // change property depending on state
        XcbUtil::get().changeState( this, XcbDefines::_NET_WM_STATE_STICKY, 1);

    } else if( XcbUtil::get().isSupported( XcbDefines::_NET_WM_DESKTOP ) ) {

        unsigned long desktop = XcbUtil::get().cardinal( XcbUtil::get().appRootWindow(), XcbDefines::_NET_CURRENT_DESKTOP );
        XcbUtil::get().changeCardinal( this, XcbDefines::_NET_WM_DESKTOP, state ? XcbDefines::ALL_DESKTOPS : desktop );

    }

    #endif

    return;

}

//________________________________________________
void TaskBarBase::_toggleRoundCorners( bool state )
{
    Debug::Throw( "TaskBarBase::_toggleRoundCorners.\n" );
    if( state ) _roundCorners();
    else clearMask();
}

//___________________________________________________
void TaskBarBase::_empty( QPaintDevice& device, const QRect& clip )
{

    Debug::Throw( "TaskBarBase::_Empty.\n" );

    // create painter on foreground pixmap
    QPainter painter( &device );
    painter.setClipRect( clip );

    painter.setRenderHints(QPainter::SmoothPixmapTransform);
    painter.setCompositionMode(QPainter::CompositionMode_SourceOver);

    // draw background pixmap into entire window
    painter.drawPixmap( 0, 0, _backgroundPixmap() );

    const QRect innerRect( _margins().adjustedRect( this->rect() ) );
    QPixmap pixmap( innerRect.size() );
    pixmap.fill( Qt::transparent );
    {
        QPainter painter( &pixmap );
        painter.setPen( foregroundColor() );
        painter.drawText( pixmap.rect(), Qt::AlignCenter|Qt::TextWordWrap, _emptyText() );
    }

    if( shadowOffset() > 0 && shadowColor().isValid() )
    {
        QImage image( pixmap.toImage() );
        Transparency::Effects::shadowBlur(image, shadowOffset(), shadowColor() );
        painter.drawImage( innerRect.topLeft(), image);
    }

    painter.drawPixmap( innerRect.topLeft(), pixmap );

    painter.end();

}

//___________________________________________________
void TaskBarBase::_updatePosition( void )
{

    Debug::Throw( "TaskBarBase::_updatePosition.\n" );
    if( attributes_.positionMode() == TaskBarAttributes::PositionAny ) return;

    // update position to account for anchor
    QDesktopWidget* desktop( qApp->desktop() );
    const QRect desktopGeometry( desktop->availableGeometry( desktop->screenNumber( this ) ) );

    // this widget size
    const QSize& size( TaskBarBase::size() );

    QPoint point( desktopGeometry.topLeft() );
    switch( attributes_.positionMode() )
    {
        case TaskBarAttributes::PositionTopLeft:
        break;

        case TaskBarAttributes::PositionTopCenter:
        point.rx()+= ( desktopGeometry.width() - size.width() )/2;
        break;

        case TaskBarAttributes::PositionTopRight:
        point.rx()+= ( desktopGeometry.width() - size.width() );
        break;

        case TaskBarAttributes::PositionBottomLeft:
        point.ry() += ( desktopGeometry.height() - size.height() );
        break;

        case TaskBarAttributes::PositionBottomCenter:
        point += QPoint( ( desktopGeometry.width() - size.width() )/2, desktopGeometry.height() - size.height() );
        break;

        case TaskBarAttributes::PositionBottomRight:
        point += QPoint( desktopGeometry.width() - size.width(), desktopGeometry.height() - size.height() );
        break;

        case TaskBarAttributes::PositionLeftCenter:
        point.ry() += ( desktopGeometry.height() - size.height() )/2;
        break;

        case TaskBarAttributes::PositionRightCenter:
        point += QPoint( desktopGeometry.width() - size.width()-1, ( desktopGeometry.height() - size.height() )/2 );
        break;

        default:
        break;
    }

    attributes_.setPosition( point );

    return;
}

//___________________________________________________
QRect TaskBarBase::_toolTipIndexRect( TaskButton::Pointer button ) const
{

    // display geometry
    const QRect displayGeometry( QRect( QPoint(0,0), frameGeometry().size() ) );

    // button geometry
    QRect buttonGeometry( button->rect() );

    const TaskBarAttributes::Layout& layout( attributes_.layout() );
    if( layout & (TaskBarAttributes::Box | TaskBarAttributes::Horizontal) )
    {

        buttonGeometry.setTop(  displayGeometry.top() );
        buttonGeometry.setBottom(  displayGeometry.bottom() );

    } else if( layout & TaskBarAttributes::Vertical ) {

        buttonGeometry.setLeft(  displayGeometry.left() );
        buttonGeometry.setRight(  displayGeometry.right() );

    }

    return buttonGeometry;

}

//____________________________________
TaskButton::Pointer TaskBarBase::_button( const QPoint& point )
{
    Debug::Throw( "TaskBarBase::_button.\n" );
    auto iter( std::find_if( buttons_.begin(), buttons_.end(), TaskButton::ContainsFTor( point ) ) );
    return (iter == buttons_.end()) ? TaskButton::null() : *iter;
}

//____________________________________
void TaskBarBase::_setCursor( const Qt::MouseButton& state )
{

    if( state == Qt::LeftButton && _moveEnabled() )  setCursor( Qt::SizeAllCursor );
    else if( state == Qt::MidButton && !attributes_.hasFlag( TaskBarAttributes::LockButtons ) )
    {

        if( attributes_.layout() & TaskBarAttributes::Horizontal ) setCursor( Qt::SizeHorCursor );
        else if( attributes_.layout() & TaskBarAttributes::Vertical ) setCursor( Qt::SizeVerCursor );
        else setCursor( Qt::SizeAllCursor );

    } else unsetCursor();

}

//____________________________________________________________
void TaskBarBase::_setAction( const TaskBarBase::Action& action )
{
    Debug::Throw() << "TaskBarBase::_setAction - action: " << action << endl;
    action_ = action;
}

//___________________________________________________
bool TaskBarBase::_updateGeometry( TaskBarBase::Action action )
{
    Debug::Throw( "TaskBarBase::_updateGeometry.\n" );

    if( !attributes_.hasFlag( TaskBarAttributes::AutoHide ) )
    {
        _setAction( NoAction );
        return false;
    }

    if( action == NoAction ) return false;

    // by default reset action to NoAction
    _setAction( NoAction );

    // window is to be minimized
    if( action == Minimize )
    {
        _setOffset( QPoint() );
        return false;
    }


    QRect mask( _minimumRect() );
    RoundedRegion::Corners corners( 0 );

    if( (attributes_.positionMode()&TaskBarAttributes::PositionBottom) && attributes_.layout() == TaskBarAttributes::HorizontalBottom )
    {

        // Bottom TaskBar
        if(
            ( action == Enter && _offset().y() <= 0 ) ||
            ( action == Leave && _offset().y() >= mask.height()-1 ) ) return false;
        _offset().ry() += ( action == Enter ) ? -1:1;
        mask.setTop( qMin( mask.top() + _offset().y(), mask.bottom()-2 ) );
        corners = RoundedRegion::TopLeft|RoundedRegion::TopRight;

    } else if( ( attributes_.positionMode()&TaskBarAttributes::PositionTop ) && attributes_.layout() ==  TaskBarAttributes::HorizontalTop ) {

        // Top TaskBar
        if(
            ( action == Enter && _offset().y() >= 0 ) ||
            ( action == Leave && _offset().y() + mask.height() <= 1 ) ) return false;

        _offset().ry() += ( action == Enter ) ? 1:-1;
        mask.setBottom( qMax( mask.bottom() + _offset().y(), mask.top()+2 ) );
        corners = RoundedRegion::BottomLeft|RoundedRegion::BottomRight;

    } else if ( (attributes_.positionMode()&TaskBarAttributes::PositionLeft) && attributes_.layout() ==  TaskBarAttributes::VerticalLeft ) {

        // Left TaskBar
        if(
            ( action == Enter && _offset().x() >= 0 ) ||
            ( action == Leave && _offset().x() + mask.width() <= 1 ) ) return false;
        _offset().rx() += ( action == Enter ) ? 1:-1;
        mask.setRight( qMax( mask.right() + _offset().x(), mask.left()+2) );
        corners = RoundedRegion::TopRight|RoundedRegion::BottomRight;

    } else if( (attributes_.positionMode()&TaskBarAttributes::PositionRight) && attributes_.layout() ==  TaskBarAttributes::VerticalRight ) {

        // Right TaskBar
        if(
            ( action == Enter && _offset().x() <= 0 ) ||
            ( action == Leave && _offset().x() >= mask.width()-1 ) ) return false;

        _offset().rx() += ( action == Enter ) ? -1:1;
        mask.setLeft( qMin( mask.left() + _offset().x(), mask.right()-2) );
        corners = RoundedRegion::TopLeft|RoundedRegion::BottomLeft;

    } else {

        return false;

    }

    // action is left to its current value
    // indicating that further updates are needed
    _setNeedUpdate( true );
    _setAction( action );
    setMask( mask, corners );

    return true;

}

//____________________________________
void TaskBarBase::_installActions( void )
{
    Debug::Throw( "TaskBarBase::_installAction" );


    reloadBackgroundAction_ = new QAction( IconEngine::get( IconNames::Reload ), tr( "Reload Background" ), this );
    reloadBackgroundAction_->setToolTip( tr( "Reload window background" ) );
    reloadBackgroundAction_->setShortcut( QKeySequence::Refresh );
    connect( reloadBackgroundAction_, SIGNAL(triggered()), &Svg::SvgEngine::get(), SLOT(forceReload()) );

    #if HAVE_XCB
    // also connect to blur background
    connect( reloadBackgroundAction_, SIGNAL(triggered()), &reloadBlurRegionAction(), SLOT(trigger()) );
    #endif


    addAction( addButtonAction_ = new QAction( IconEngine::get( IconNames::Add ), tr( "Add Button" ), this ) );
    connect( addButtonAction_, SIGNAL(triggered()), SLOT(_addButton()) );

    addAction( addSeparatorAction_ = new QAction( IconEngine::get( IconNames::Add ), tr( "Add Separator" ), this ) );
    connect( addSeparatorAction_, SIGNAL(triggered()), SLOT(_addSeparator()) );

    addAction( removeButtonAction_ = new QAction( IconEngine::get( IconNames::Remove ), tr( "Remove Button" ), this ) );
    connect( removeButtonAction_, SIGNAL(triggered()), SLOT(_removeButton()) );

    addAction( editButtonAction_ = new QAction( IconEngine::get( IconNames::Configure ), tr( "Edit Button" ), this ) );
    connect( editButtonAction_, SIGNAL(triggered()), SLOT(_editButton()) );

    addAction( editAttributesAction_ = new QAction( IconEngine::get( IconNames::Configure ), tr( "Edit Taskbar Attributes" ), this ) );
    connect( editAttributesAction_, SIGNAL(triggered()), SLOT(_editAttributes()) );

    addAction( editTaskListAction_ = new QAction( IconEngine::get( IconNames::Configure ), tr( "Edit Button List" ), this ) );
    connect( editTaskListAction_, SIGNAL(triggered()), SLOT(_editTaskList()) );

    addAction( toggleTooltipAction_ = new QAction( tr( "Show Tooltips" ), this ) );
    toggleTooltipAction_->setCheckable( true );
    connect( toggleTooltipAction_, SIGNAL(toggled(bool)), SLOT(_toggleToolTip(bool)) );

    QActionGroup* action_group = new QActionGroup( this );
    action_group->setExclusive( true );
    connect( action_group, SIGNAL(triggered(QAction*)), SLOT(_toggleHighlight(QAction*)) );

    addAction( noHighlightAction_ = new QAction( tr( "No Highlight" ), this ) );
    noHighlightAction_->setCheckable( true );
    action_group->addAction(noHighlightAction_ );

    addAction( highlightOnEnterAction_ = new QAction( tr( "Highlight on Enter" ), this ) );
    highlightOnEnterAction_->setCheckable( true );
    action_group->addAction(highlightOnEnterAction_ );

    addAction( highlightOnClickAction_ = new QAction( tr( "Highlight on Click" ), this ) );
    highlightOnClickAction_->setCheckable( true );
    action_group->addAction(highlightOnClickAction_ );

    // flag action group
    QActionGroup* group( new QActionGroup( this ) );
    group->setExclusive( false );

    // lock position
    QAction* action;
    addAction( action = new QAction( tr( "Lock Position" ), this ) );
    action->setToolTip( tr( "Lock taskbar position" ) );
    action->setCheckable( true );
    action->setChecked( false );
    group->addAction( action );
    flagActions_.insert( action, TaskBarAttributes::LockPosition );

    // lock position
    addAction( action = new QAction( tr( "Auto Hide" ), this ) );
    action->setToolTip( tr( "Auto hide taskbar" ) );
    action->setCheckable( true );
    action->setChecked( false );
    group->addAction( action );
    flagActions_.insert( action, TaskBarAttributes::AutoHide );

    // stays on top
    addAction( action = new QAction( tr( "Keep Above" ), this ) );
    action->setToolTip( tr( "Keep taskbar above all others" ) );
    action->setCheckable( true );
    group->addAction( action );
    flagActions_.insert( action, TaskBarAttributes::StaysOnTop );

    // sticky
    addAction( action = new QAction( tr( "Sticky" ), this ) );
    action->setToolTip( tr( "Make taskbar appear on all desktops" ) );
    action->setCheckable( true );
    action->setChecked( false );
    group->addAction( action );
    flagActions_.insert( action, TaskBarAttributes::Sticky );

    // hide from taskbar
    addAction( action = new QAction( tr( "Hide from Taskbar" ), this ) );
    action->setCheckable( true );
    action->setChecked( false );
    group->addAction( action );
    flagActions_.insert( action, TaskBarAttributes::HideFromTaskBar );

    // round corners
    addAction( action = new QAction( tr( "Round Corners" ), this ) );
    action->setToolTip( tr( "Round window corners" ) );
    action->setCheckable( true );
    action->setChecked( false );
    action->setEnabled( windowFlags() & Qt::FramelessWindowHint );
    group->addAction( action );
    flagActions_.insert( action, TaskBarAttributes::RoundCorners );

    // lock button position
    addAction( action = new QAction( tr( "Lock Buttons' Position" ), this ) );
    action->setToolTip( tr( "Lock buttons' position" ) );
    action->setCheckable( true );
    action->setChecked( false );
    group->addAction( action );
    flagActions_.insert( action, TaskBarAttributes::LockButtons );

    // show button name
    addAction( action = new QAction( tr( "Show Button Names" ), this ) );
    action->setToolTip( tr( "Show button names below icon" ) );
    action->setCheckable( true );
    action->setChecked( false );
    group->addAction( action );
    flagActions_.insert( action, TaskBarAttributes::ShowName );

    // expand separators
    addAction( action = new QAction( tr( "Expand Separators" ), this ) );
    action->setToolTip( tr( "Expand separators when zooming" ) );
    action->setCheckable( true );
    action->setChecked( false );
    group->addAction( action );
    flagActions_.insert( action, TaskBarAttributes::ExpandSeparators );

    connect( group, SIGNAL(triggered(QAction*)), SLOT(_toggleFlagAction(QAction*)) );

}

//________________________________________________
void TaskBarBase::_roundCorners( void )
{ setMask( rect() ); }

//________________________________________________
bool TaskBarBase::_optionsChanged( void )
{

    Debug::Throw( "TaskBarBase::_optionsChanged.\n" );
    bool out( false );
    for( auto& pair:storedOptions_ )
    {
        Option current;
        if( XmlOptions::get().contains( pair.first ) && (current = XmlOptions::get().option( pair.first )) == pair.second ) continue;

        Debug::Throw() << "TaskBarBase::_optionsChanged -"
            << " option: " << pair.first
            << " old value: " << pair.second.raw()
            << " new value: " << current.raw()
            << endl;
        pair.second = current;
        out = true;

    }

    return out;

}

//________________________________________________
void TaskBarBase::_updateTasks( const TaskBarAttributes& attributesList )
{

    Debug::Throw( "TaskBarBase::_updateTasks.\n" );

    // reset
    _setSelectedButton( TaskButton::null() );
    _clearButtons();
    for( const auto& attributes:attributesList.taskAttributesList() )
    {

        // create button, check validity and add to list
        TaskButton::Pointer button( new TaskButton( attributesList, attributes ) );
        if( button->isValid() ) buttons_ << ( button );
        else button->deleteLater();

    }

}
