#ifndef TaskBarBase_h
#define TaskBarBase_h


/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "RoundedRegion.h"
#include "TaskAttributes.h"
#include "TaskBarAttributes.h"
#include "TaskButton.h"
#include "TransparentWidget.h"

#include <QBasicTimer>
#include <QTimerEvent>
#include <QList>
#include <QHash>

class Menu;
class ToolTipWidget;

//* TaskBar main widget
class TaskBarBase: public Transparency::TransparentWidget
{

    //* Qt meta object declaration
    Q_OBJECT

    public:

    //* list
    using List = QList<TaskBarBase*>;

    //* constructor
    TaskBarBase( QWidget* = nullptr );

    //* synchronize task button attributes and taskBar attributes
    virtual void synchronizeAttributes( void );

    //* attributes
    virtual const TaskBarAttributes& attributes( void ) const
    { return attributes_; }

    //* attributes
    virtual TaskBarAttributes& attributes( void )
    { return attributes_; }

    //* attributes
    virtual void setAttributes( const TaskBarAttributes&, bool alsoUpdate = true );

    //* mask
    virtual void setMask( QRect );

    //* mask
    virtual void setMask( QRect, RoundedRegion::Corners );

    //* empty size
    QSize emptySize( void ) const;

    //* used to retrieve a taskBar from attributes
    class SameAttributesFTor
    {

        public:

        //* constructor
        SameAttributesFTor( const TaskBarAttributes& attributes ):
            attributes_( attributes )
            {}

        //* predicate
        bool operator()( TaskBarBase* task ) const
        { return task->attributes().id() == attributes_.id(); }

        private:

        //* predicted attributes
        const TaskBarAttributes& attributes_;

    };

    //*@name actions
    //@{

    //* add button
    QAction& addButtonAction( void ) const
    { return *addButtonAction_; }

    //* add separator
    QAction& addSeparatorAction( void ) const
    { return *addSeparatorAction_; }

    //* remove button action
    QAction& removeButtonAction( void ) const
    { return *removeButtonAction_; }

    //* edit button
    QAction& editButtonAction( void ) const
    { return *editButtonAction_; }

    //* edit taskBar attributes
    QAction& editAttributesAction( void ) const
    { return *editAttributesAction_; }

    //* edit taskBar button list
    QAction& editTaskListAction( void ) const
    { return *editTaskListAction_; }

    //* toggle tooltip
    QAction& toggleToolTipAction( void ) const
    { return *toggleTooltipAction_; }

    //* no highlight
    QAction& noHighlightAction( void ) const
    { return *noHighlightAction_; }

    //* highlight on click
    QAction& highlightOnClickAction( void ) const
    { return *highlightOnClickAction_; }

    //* highlight on enter
    QAction& highlightOnEnterAction( void ) const
    { return *highlightOnEnterAction_; }

    //* reload background
    QAction& reloadBackgroundAction( void ) const
    { return *reloadBackgroundAction_; }

    //* map actions to TaskBarAttributes flags
    using ActionMap = QHash<QAction*, TaskBarAttributes::Flag>;

    //* map actions to TaskBarAttributes flags
    const ActionMap& flagActions( void ) const
    { return flagActions_; }

    //* map actions to TaskBarAttributes flags
    QAction* flagAction( const TaskBarAttributes::Flag& flag ) const
    {
        for( ActionMap::const_iterator iter = flagActions().begin(); iter != flagActions().end(); iter++ )
        { if( iter.value() == flag ) return iter.key(); }

        return 0L;

    }

    //@}

    Q_SIGNALS:

    //* taskBar modified
    void modified( void );

    //* emited when attributes are changed
    void attributesChanged( const TaskBarAttributes& );

    //* task list changed
    void taskListChanged( void );

    //* taskBar entered
    /** this is used so that Application knows which is the current taskBar */
    void taskBarEntered( const TaskBarAttributes& );

    public Q_SLOTS:

    //* update configuration
    virtual void updateConfiguration( void );

    protected Q_SLOTS:

    //* add button
    virtual void _addButton( bool is_separator = false );

    //* add separator
    virtual void _addSeparator( void )
    { _addButton( true ); }

    //* remove button
    virtual void _removeButton( void );

    //* edit button
    virtual void _editButton( void );

    //* edit attributes
    virtual void _editAttributes( void );

    //* edit attributes
    virtual void _editTaskList( void );

    //* enable move
    virtual void _setEnableMove( bool value = true );

    //* toggle action associated to TaskBarAttributes flag
    virtual void _toggleFlagAction( QAction* );

    //* toggle highlight
    virtual void _toggleHighlight( QAction* );

    //* toggle tooltip
    virtual void _toggleToolTip( bool );

    //*  change taskBar position when desktop is resized
    virtual void _desktopResizedDelayed( void );

    //*  change taskBar position when desktop is resized
    virtual void _desktopResized( void );

    protected:

    //*@name event handlers
    //@{

    #if HAVE_XCB && QT_VERSION < 0x050000
    //* check X11 event
    bool x11Event( XEvent* event );
    #endif

    //* timer event, to toggle repaint
    virtual void timerEvent( QTimerEvent* );

    //* resize event method
    virtual void resizeEvent( QResizeEvent* );

    //* pointer enters window
    virtual void enterEvent( QEvent* );

    //* pointer leaves window
    virtual void leaveEvent( QEvent* );

    //* mouse move event [overloaded]
    virtual void mousePressEvent( QMouseEvent* );

    //* mouse move event [overloaded]
    virtual void mouseReleaseEvent( QMouseEvent* );

    //* mouse move event [overloaded]
    virtual void mouseMoveEvent( QMouseEvent* );

    //* context menu event [overloaded]
    virtual void contextMenuEvent( QContextMenuEvent* );

    //* paint window
    virtual void _paint( QPaintDevice&, const QRect& );

    //@}

    //* attributes
    virtual void _resetAttributes( void )
    { setAttributes( attributes_ ); }

    //* default context menu
    virtual Menu* _createContextMenu( void );

    //* buttons
    virtual TaskButton::List& _buttons( void )
    { return buttons_; }

    //* buttons
    virtual const TaskButton::List& _buttons( void ) const
    { return buttons_; }

    //* Save TaskBarAttributess; emit Modified signal is argument is true
    virtual void _saveButtons( void );

    //* clear buttons
    virtual void _clearButtons( void );

    //* selected button
    virtual TaskButton::Pointer _selectedButton( void )
    { return selectedButton_; }

    //* selected button
    virtual void _setSelectedButton( TaskButton::Pointer button )
    { selectedButton_ = button; }

    //* number of columns
    virtual const int& _numColumns( void ) const
    { return nColumns_; }

    //* number of lines
    virtual const int& _numLines( void ) const
    { return nLines_; }

    //* set dimensions
    virtual void _setDimensions( void );

    // true if taskBar is being moved
    bool _moveEnabled( void ) const
    { return enableMove_; }

    //* moved
    virtual void _setMoved( bool value )
    { moved_ = value; }

    //* moved
    virtual bool _moved( void ) const
    { return moved_; }

    //* need update
    virtual void _setNeedUpdate( bool value )
    { needUpdate_ = value; }

    //* need update
    bool _needUpdate( void ) const
    { return needUpdate_; }

    //* button state
    virtual const Qt::MouseButton& _mouseButton( void ) const
    { return mouseButton_; }

    //* button state
    virtual void _setMouseButton( const Qt::MouseButton& state )
    { mouseButton_ = state; }

    //* spacing
    virtual int _spacing( void ) const
    { return spacing_; }

    //* Step
    virtual int _step( void ) const
    { return step_; }

    //* click position
    virtual const QPoint& _clickPosition( void ) const
    { return clickPos_; }

    //* click position
    virtual void _setClickPosition( const QPoint& point )
    { clickPos_ = point; }

    //* mouse position
    virtual const QPoint& _mousePosition( void ) const
    { return mousePos_; }

    //* click position
    virtual void _setMousePosition( const QPoint& point )
    { mousePos_ = point; }

    //*@name timers
    //@{

    //* update timer
    QBasicTimer& _updateTimer( void )
    { return updateTimer_; }

    //* geometry timer
    QBasicTimer& _geometryTimer( void )
    { return geometryTimer_; }

    //* update timer
    QBasicTimer& _enableMoveTimer( void )
    { return enableMoveTimer_; }

    //@}

    //* context_menu
    Menu& _contextMenu( void )
    {
        if( !contextMenu_ ) contextMenu_ = _createContextMenu();
        return *contextMenu_;
    }

    //* deselect button
    virtual void _enterButton( TaskButton::Pointer button );

    //* deselect button
    virtual void _leaveButton( TaskButton::Pointer button );

    //* press button
    virtual void _pressButton( TaskButton::Pointer button );

    //* release button
    virtual void _releaseButton( TaskButton::Pointer button );

    //* swap buttons
    void _swap( TaskButton::Pointer first, TaskButton::Pointer second );

    //* stays on top
    virtual void _toggleStaysOnTop( bool );

    //* toggle hide from taskbar
    virtual void _toggleHideFromTaskBar( bool );

    //* toggle window stickyness
    virtual void _toggleSticky( bool );

    //* round corners
    virtual void _toggleRoundCorners( bool );

    //* display some text in case the taskBar is empty
    virtual void _empty( QPaintDevice&, const QRect& );

    //* resize the taskBar so that is is able to contain all buttons
    virtual void _resize( void ) = 0;

    //* place buttons in the taskBar
    virtual void _placeButtons( void ) = 0;

    //* draw buttons
    /**
    if force is true, only buttons which are zooming are drawn
    returns true if at least one if zooming
    */
    virtual bool _drawButtons( QPaintDevice&, QRect ) = 0;

    //* update attributes position from predefined values
    virtual void _updatePosition( void );

    //* return tooltip item rect for a given button
    QRect _toolTipIndexRect( TaskButton::Pointer ) const;

    //* return task button matching position; 0 if no match
    virtual TaskButton::Pointer _button( const QPoint& point );

    //* set cursor according to button state
    void _setCursor( const Qt::MouseButton& state );

    //* enter/leave action.
    /** Needed to update geometry accordingly */
    enum Action
    {
        NoAction,
        Enter,
        Leave,
        Minimize
    };

    //* enter/leave action
    void _setAction( const Action& action );

    //* enter/leave action
    const Action& _action( void ) const
    { return action_; }

    //* minimum size
    virtual QSize _minimumSize() const
    { return size(); }

    //* minimum size
    virtual QRect _minimumRect() const
    { return QRect( QPoint(0,0), size() ); }

    //* update geometry (mask) for enter/leave event
    virtual bool _updateGeometry( Action action );

    //* offset (for autoHide)
    virtual const QPoint& _offset( void ) const
    { return offset_; }

    //* offset (for autoHide)
    virtual QPoint& _offset( void )
    { return offset_; }

    //* offset (for autoHide)
    virtual void _setOffset( const QPoint& point )
    { offset_ = point; }

    //* use svg
    virtual bool _useSvg( void ) const
    { return useSvg_; }

    //* use svg
    virtual void _setUseSvg( bool value )
    { useSvg_ = value; }

    //* round corners by means of a mask
    void _roundCorners( void );

    //* empty text
    QString _emptyText( void ) const
    { return tr( "The taskbar is empty. Right click to add Buttons." ); }

    //* disable input shape
    virtual void _updateInputShape( void )
    { return; }

    private:

    //* actions
    void _installActions( void );

    //* true if some of the stored options have changed
    bool _optionsChanged( void );

    //* update task buttons
    void _updateTasks( const TaskBarAttributes& attributes );

    //*@name attributes
    //@{

    //* taskBar attributes
    TaskBarAttributes attributes_;

    //* step for window displacement
    int step_ = 1;

    //* spacing between buttons
    int spacing_ = 0;

    //* number of columns
    int nColumns_ = 0;

    //* number of lines
    int nLines_ = 0;

    //@}

    //*@name displacements
    //@{

    //* click position
    QPoint clickPos_;

    //* current mouse position
    QPoint mousePos_;

    //* button state
    Qt::MouseButton mouseButton_ = Qt::NoButton;

    //* timer to enable taskBar moving after left click
    QBasicTimer enableMoveTimer_;

    //* set to true when mouse pointer is allowed to move window (use timer
    bool enableMove_ = false;

    //* is set to true if taskBar was moved between press and release
    bool moved_ = false;

    //@}

    //* need button update
    bool needUpdate_ = true;

    //* menu
    Menu* contextMenu_ = nullptr;

    //* running tooltip
    ToolTipWidget* toolTipWidget_ = nullptr;

    //* enter/leave action
    Action action_ = NoAction;

    //* offset (for autoHide)
    QPoint offset_;

    //* use svg
    bool useSvg_ = true;

    //*@name actions
    //@{

    //* reload background
    QAction* reloadBackgroundAction_ = nullptr;

    //* add button
    QAction* addButtonAction_ = nullptr;

    //* add separator
    QAction* addSeparatorAction_ = nullptr;

    //* remove button
    QAction* removeButtonAction_ = nullptr;

    //* edit button
    QAction* editButtonAction_ = nullptr;

    //* edit attributes
    QAction* editAttributesAction_ = nullptr;

    //* edit attributes
    QAction* editTaskListAction_ = nullptr;

    //* toggle tooltip
    QAction* toggleTooltipAction_ = nullptr;

    //* map actions to TaskBarAttributes flags
    ActionMap flagActions_;

    //* no button highlight
    QAction* noHighlightAction_ = nullptr;

    //* highlight buttons on enter
    QAction* highlightOnEnterAction_ = nullptr;

    //* highlight buttons on click
    QAction* highlightOnClickAction_ = nullptr;

    //@}

    //* options whose modifications are to be checked, to decide if tasks must be reset
    QList< Options::Pair > storedOptions_;

    //*@name buttons
    //@{

    //* Selected button
    TaskButton::Pointer selectedButton_;

    //* update timer
    QBasicTimer updateTimer_;

    //* geometry update timer
    QBasicTimer geometryTimer_;

    //* stored list
    TaskButton::List buttons_;

    //@}

};

#endif
