/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "TaskBarMinimum.h"

#include "CompositeEngine.h"
#include "RoundedRegion.h"
#include "SvgEngine.h"
#include "ToolTipWidget.h"
#include "XmlOptions.h"

#include <QAction>
#include <QApplication>
#include <QDesktopWidget>

//____________________________________________________
TaskBarMinimum::TaskBarMinimum( QWidget* parent ):
    TaskBarBase( parent ),
    rootPixmapChanged_( true )
{ Debug::Throw( "TaskBarMinimum::TaskBarMinimum.\n" ); }

//____________________________________________________
void TaskBarMinimum::_updateRootPixmap( void )
{

    Debug::Throw( "TaskBarMinimum::_updateRootPixmap" );
    _setRootPixmapChanged( false );
    if( Debug::level() >= 1 )
    {
        // solid background
        _setRootPixmap( QPixmap( size() ) );
        _rootPixmap().fill( palette().color( backgroundRole() ) );

    } else if( Transparency::CompositeEngine::get().isAvailable() ) {

        _setRootPixmap( QPixmap( size() ) );
        QPainter painter( &_rootPixmap() );
        painter.setRenderHints(QPainter::SmoothPixmapTransform);
        painter.setCompositionMode(QPainter::CompositionMode_Source );
        painter.fillRect( _rootPixmap().rect(), Qt::transparent);

    } else {

        _setRootPixmap( QPixmap::grabWindow(
            QApplication::desktop()->winId(),
            geometry().x(), geometry().y(),
            geometry().width(), geometry().height() ) );

    }
}

//___________________________________________________
void TaskBarMinimum::_updateBackgroundPixmap( void )
{
    Debug::Throw( "TaskBarMinimum::TaskBarFixed.\n" );
    if( _rootPixmapChanged() )
    {
        // there is no need to update the root pixmap when composite engine is off,
        // unless during resize events, for which the method is called directly
        if( !Transparency::CompositeEngine::get().isAvailable() ) _updateRootPixmap();
        else _setRootPixmapChanged( false );
    }

    TaskBarBase::_updateBackgroundPixmap();

    // need to draw svg on background pixmap if button list is empty
    // due to the way empty taskBars are implemented
    if( _buttons().empty() &&_useSvg() && Svg::SvgEngine::get().isValid() )
    {
        QPainter painter( &_backgroundPixmap() );
        painter.setRenderHints(QPainter::SmoothPixmapTransform);
        painter.setCompositionMode(QPainter::CompositionMode_Source );
        painter.drawPixmap( QPoint(0,0),  Svg::SvgEngine::get().get( size() ) );
    }

}

//______________________________________________
void TaskBarMinimum::resizeEvent( QResizeEvent* event )
{
    _updateRootPixmap();

    /*
    need to bypass TaskBarBase::resizeEvent, in order not to have
    the entire background blurred, when enabled
    */
    if( attributes().hasFlag( TaskBarAttributes::RoundCorners ) ) _roundCorners();
    TransparentWidget::resizeEvent( event );
}

//______________________________________________________
void TaskBarMinimum::_resize( void )
{

    Debug::Throw( "TaskBarMinimum::_resize.\n" );

    // if no tasks, set empty message and exit
    if( _buttons().empty() ) {
        resize( emptySize() );
        return;
    }

    // dimensions
    _setDimensions();
    _updateMinimumSize();
    _updateMaximumSize();
    _updateMinimumRectPosition();

    // svg
    if( _useSvg() && Svg::SvgEngine::get().isValid() ) _preloadSvg();

    // make sure taskBar is in minimum state
    // resize
    _setRootPixmapChanged( true );
    _setAction( Minimize );
    resize( _maximumSize() );

}

//___________________________________________________
void TaskBarMinimum::_placeButtons( void )
{

    Debug::Throw( "TaskBarMinimum::_placeButtons.\n" );

    // check that there is at least one button
    if( _buttons().empty() ) return;

    // starting positions
    int x(0);
    int y(0);

    const TaskBarAttributes::Layout &layout( attributes().layout() );

    if( layout & ( TaskBarAttributes::Vertical | TaskBarAttributes::Horizontal ) )
    {

        for( const auto& button:_buttons() )
        {

            button->setRect( QPoint(x,y), QSize( button->currentSize().width() + _spacing(), button->currentSize().height() + _spacing()  ) );

            // update to next button
            if( layout & TaskBarAttributes::Vertical ) y+= button->rect().height();
            else if( layout & TaskBarAttributes::Horizontal )  x += button->rect().width();
        }

    } else if( layout & TaskBarAttributes::Box ) {

        int columnIndex(0);
        int lineIndex(0);

        // compute column width/line height
        QVector<int> columnWidth( _numColumns(), 0 );
        QVector<int> lineHeight( _numLines(), 0 );
        for( const auto& button:_buttons() )
        {

            lineHeight[lineIndex] = qMax( lineHeight[lineIndex], button->currentSize().height() + _spacing() );
            columnWidth[columnIndex] = qMax( columnWidth[columnIndex], button->currentSize().width() + _spacing() );
            columnIndex++;
            if( columnIndex >= _numColumns() ) {
                columnIndex = 0;
                lineIndex++;
            }

        }

        // place buttons
        columnIndex = 0;
        lineIndex = 0;
        for( const auto& button:_buttons() )
        {
            button->setRect( QPoint( x, y ), QSize( columnWidth[columnIndex], lineHeight[lineIndex] ) );
            x+=columnWidth[columnIndex];
            columnIndex++;
            if( columnIndex >= _numColumns() ) {
                x = 0;
                columnIndex = 0;

                y+= lineHeight[lineIndex];
                lineIndex++;
            }
        }
    }

}

//______________________________________________________
void TaskBarMinimum::_addOffsets( void )
{

    // dump
    Debug::Throw() << "TaskBarMinimum::_addOffsets -"
        << " size: (" << width() << "," << height() << ")"
        << endl;

    // get bounding rect for current button sizes
    QRect boundingRect;
    for( const auto& button:_buttons() )
    {
        boundingRect.setWidth( qMax( boundingRect.width(), button->rect().right() ) );
        boundingRect.setHeight( qMax( boundingRect.height(), button->rect().bottom() ) );
    }

    // compute offsets
    QPoint offset( (size().width() - boundingRect.width())/2, (size().height() - boundingRect.height())/2 );
    const TaskBarAttributes::Layout &layout( attributes().layout() );

    // add offsets
    for( const auto& button:_buttons() )
    {
        QPoint buttonOffset( offset );
        if( layout & TaskBarAttributes::Vertical )
        {

            if( layout & TaskBarAttributes::TopLeft ) buttonOffset.setX( _margins().left() );
            else if( layout & TaskBarAttributes::BottomRight ) buttonOffset.setX( width() - button->rect().width() - _margins().right() );
            else if( layout & TaskBarAttributes::Center ) buttonOffset.setX( (width() - button->rect().width())/2 );

        } else if( layout & TaskBarAttributes::Horizontal ) {

            if( layout & TaskBarAttributes::TopLeft ) buttonOffset.setY( _margins().top() );
            else if( layout & TaskBarAttributes::BottomRight ) buttonOffset.setY( height() - button->rect().height() - _margins().bottom() );
            else if( layout & TaskBarAttributes::Center ) buttonOffset.setY( (height() - button->rect().height())/2 );

        }

        button->setRect( button->rect().translated( buttonOffset ) );

    }

}

//___________________________________________________
bool TaskBarMinimum::_drawButtons( QPaintDevice& device, QRect rect )
{

    // if no tasks, do nothing
    if( _buttons().empty() ) return false;

    // place the buttons
    _placeButtons();
    _addOffsets();

    // redraw
    QPainter painter( &device );
    painter.setClipRect( rect );

    if( Transparency::CompositeEngine::get().isAvailable() )
    {
        painter.setRenderHints(QPainter::SmoothPixmapTransform);
        painter.setCompositionMode(QPainter::CompositionMode_Source );
    }

    // draw root pixmap into entire window
    painter.drawPixmap( 0, 0, _rootPixmap() );

    // draw background pixmap clipped to visibleRect
    _updateVisibleRect();
    _drawVisibleRect( painter );

    // draw buttons, translated by offset if needed
    painter.setCompositionMode(QPainter::CompositionMode_SourceOver );
    painter.translate( _offset() );

    bool out( false );
    for( const auto& button:_buttons() )
    {

        button->paint( painter );
        if( _action() != Enter )
        {
            button->updateDirection();
            button->updateHighlightDirection();
            out |= (button->direction() != TaskButton::NoDirection);
            out |= button->highlightDirection() != TaskButton::NoHighlight;
        }

    }

    painter.end();

    return out;

}

//___________________________________________________
void TaskBarMinimum::_drawVisibleRect( QPainter& painter )
{
    painter.save();

    if( _useSvg() && Svg::SvgEngine::get().isValid() )
    {

        painter.translate( _visibleRect().topLeft() );
        painter.drawPixmap( QPoint(0,0), Svg::SvgEngine::get().get( _visibleRect().size() ) );

    } else {
        if( attributes().hasFlag( TaskBarAttributes::RoundCorners ) ) painter.setClipRegion( RoundedRegion( _visibleRect(), RoundedRegion::All ) );
        else painter.setClipRegion( _visibleRect() );
        painter.drawPixmap( 0, 0, _backgroundPixmap() );
    }
    painter.restore();
}

//_____________________________________________________________
void TaskBarMinimum::_updateMinimumSize( void )
{

    Debug::Throw( "TaskBarMinimum::_updateMinimumSize.\n" );

    // compute minimum size
    QSize minimumSize;

    // compute visible rect
    const TaskBarAttributes::Layout& layout ( attributes().layout() );
    if( layout & TaskBarAttributes::Vertical )
    {
        for( const auto& button:_buttons() )
        {
            minimumSize.setHeight( minimumSize.height() + button->minimumSize().height() + _spacing() );
            minimumSize.setWidth( qMax<int>( minimumSize.width(), button->minimumSize().width() + _spacing() + _margins().width() ) );
        }

        minimumSize.setHeight( minimumSize.height() + _margins().height() );

    } else if( layout & TaskBarAttributes::Horizontal ) {

        for( const auto& button:_buttons() )
        {
            minimumSize.setWidth( minimumSize.width() + button->minimumSize().width() + _spacing() );
            minimumSize.setHeight( qMax<int>( minimumSize.height(), button->minimumSize().height() + _spacing() + _margins().height() ) );
        }

        minimumSize.setWidth( minimumSize.width() + _margins().width() );

    } else {

        int columnIndex(0);
        int lineIndex(0);

        // compute column width/line height
        QVector<int> columnWidth( _numColumns(), 0 );
        QVector<int> lineHeight( _numLines(), 0 );
        for( const auto& button:_buttons() )
        {

            lineHeight[lineIndex] = qMax( lineHeight[lineIndex], button->currentSize().height() + _spacing() );
            columnWidth[columnIndex] = qMax( columnWidth[columnIndex], button->currentSize().width() + _spacing() );
            columnIndex++;
            if( columnIndex >= _numColumns() ) {
                columnIndex = 0;
                lineIndex++;
            }

        }

        int lineWidth(0);
        columnIndex = 0;
        lineIndex = 0;
        for( const auto& button:_buttons() )
        {

            Q_UNUSED( button );
            lineWidth += columnWidth[columnIndex];
            columnIndex++;

            // change column
            if( columnIndex >= _numColumns() )
            {
                minimumSize.setWidth( qMax( minimumSize.width(), lineWidth ) );
                minimumSize.setHeight( minimumSize.height()+lineHeight[lineIndex] );

                lineWidth = 0;
                columnIndex = 0;
                lineIndex++;

            }

        }

        minimumSize.setHeight( minimumSize.height() + lineHeight.back() + _margins().height() );
        minimumSize.setWidth( minimumSize.width() + _margins().width() );

    }

    // update
    _setMinimumSize( minimumSize );
    return;

}

//_____________________________________________________________
void TaskBarMinimum::_updateMaximumSize( void )
{

    Debug::Throw( "TaskBarMinimum::_updateMaximumSize.\n" );

    // compute maximum size
    const QSize& minimumSize( _minimumSize() );
    QSize maximumSize( _minimumSize() );
    for( const auto& button:_buttons() )
    {
        maximumSize.setWidth( qMax( maximumSize.width(), minimumSize.width() + button->maximumSize().width() - button->minimumSize().width() ) );
        maximumSize.setHeight( qMax( maximumSize.height(), minimumSize.height() + button->maximumSize().height() - button->minimumSize().height() ) );
    }

    // update
    _setMaximumSize( maximumSize );
    return;

}

//_____________________________________________________________
void TaskBarMinimum::_updateMinimumRectPosition( void )
{

    Debug::Throw( "TaskBarMinimum::_updateMinimumRectPosition.\n" );

    QRect minimumRect( _minimumRect() );
    QPoint offset;
    const TaskBarAttributes::Layout& layout ( attributes().layout() );
    if( layout & TaskBarAttributes::Vertical ) {

        offset.setY( ( _maximumSize().height() - minimumRect.height() )/2 );
        if( layout & TaskBarAttributes::BottomRight ) offset.setX( _maximumSize().width() - minimumRect.width() );
        else if( layout & TaskBarAttributes::Center ) offset.setX( ( _maximumSize().width() - minimumRect.width() )/2 );

    } else if( layout & TaskBarAttributes::Horizontal ) {
        offset.setX( ( _maximumSize().width() - minimumRect.width() )/2 );
        if( layout & TaskBarAttributes::BottomRight ) offset.setY( _maximumSize().height() - minimumRect.height() );
        else if( layout & TaskBarAttributes::Center ) offset.setY( ( _maximumSize().height() - minimumRect.height() )/2 );

    } else  {

        offset.setX( ( _maximumSize().width() - minimumRect.width() )/2 );
        offset.setY( ( _maximumSize().height() - minimumRect.height() )/2 );

    }

    minimumRect.moveTo( offset );
    _setMinimumRect( minimumRect );

}

//_____________________________________________________________
void TaskBarMinimum::_updateVisibleRect( void )
{

    Debug::Throw( "TaskBarMinimum::_updateVisibleRect.\n" );

    QRect visibleRect;
    const TaskBarAttributes::Layout& layout ( attributes().layout() );
    if( layout & TaskBarAttributes::Vertical )
    {

        // width is from minimum size
        visibleRect.setWidth( _minimumSize().width() );

        // get bounding height and y offset
        bool first( true );
        for( const auto& button:_buttons() )
        {
            visibleRect.setTop( first ? button->rect().top() : qMin( visibleRect.top(), button->rect().top() ) );
            visibleRect.setBottom( qMax( visibleRect.bottom(), button->rect().bottom() ) );
            first = false;
        }

        visibleRect.setTop( visibleRect.top() - _margins().top() );
        visibleRect.setBottom( visibleRect.bottom() + _margins().bottom() );

        if( layout & TaskBarAttributes::BottomRight ) visibleRect.moveLeft(  _maximumSize().width() - visibleRect.width() );
        else if( layout & TaskBarAttributes::Center ) visibleRect.moveLeft( ( _maximumSize().width() - visibleRect.width() )/2 );

    } else if( layout & TaskBarAttributes::Horizontal ) {

        // height is from minimum size
        visibleRect.setHeight( _minimumSize().height() );

        // get bounding height and y offset
        bool first( true );
        for( const auto& button:_buttons() )
        {
            visibleRect.setLeft( first ? button->rect().left() : qMin( visibleRect.left(), button->rect().left() ) );
            visibleRect.setRight( qMax( visibleRect.right(), button->rect().right() ) );
            first = false;
        }

        visibleRect.setLeft( visibleRect.left() - _margins().left() );
        visibleRect.setRight( visibleRect.right() + _margins().right() );

        if( layout & TaskBarAttributes::BottomRight ) visibleRect.moveTop(  _maximumSize().height() - visibleRect.height() );
        else if( layout & TaskBarAttributes::Center ) visibleRect.moveTop( ( _maximumSize().height() - visibleRect.height() )/2 );

    } else {

        visibleRect.setSize( _minimumSize() );
        visibleRect.moveTo( ( _maximumSize().width() - visibleRect.width() )/2, ( _maximumSize().height() - visibleRect.height() )/2 );

    }

    // add offset (for auto-hide)
    visibleRect.translate( _offset() );

    // in case of autoHide, one make sure there is at least
    // a thin band drawn at the bottom, that match the hidden geometry
    if( attributes().hasFlag( TaskBarAttributes::AutoHide) )
    {

        const TaskBarAttributes::Layout& layout ( attributes().layout() );
        if( layout & TaskBarAttributes::Horizontal )
        {
            if( layout & TaskBarAttributes::TopLeft )  visibleRect.setBottom( qMax(  visibleRect.bottom(), 3 ) );
            else visibleRect.setTop( qMin( visibleRect.top(), height()-3 ) );
        } else {
            if( layout & TaskBarAttributes::TopLeft ) visibleRect.setRight( qMax( visibleRect.right(), 3 ) );
            else visibleRect.setLeft( qMin( visibleRect.left(), width()-3 ) );
        }

    }

    // update blur areas if visibleRect is changed
    if( _setVisibleRect( visibleRect ) && Transparency::CompositeEngine::get().isAvailable() )
    { _updateBlurRegion( _outerPadding().adjustedRect( visibleRect ).translated( mapTo( window(), QPoint(0,0) ) ) ); }

    return;

}

//_____________________________________________________________
bool TaskBarMinimum::_updateGeometry( Action action )
{

    Debug::Throw() << "TaskBarMinimum::_updateGeometry - action: " << action << endl;

    if( action == NoAction ) return false;

    // minimum action
    if( action == Minimize )
    {

        setMask( _minimumRect() );
        if( !Transparency::CompositeEngine::get().isAvailable() ) _setRootPixmapChanged( true );
        _setAction( NoAction );
        return false;

    }

    // enter action
    if( action == Enter ) {

        if( _rootPixmapChanged() )
        {
            // there is no need to update the root pixmap when composite engine is off,
            // unless during resize events, for which the method is called directly
            if( !Transparency::CompositeEngine::get().isAvailable() ) _updateRootPixmap();
            else _setRootPixmapChanged( false );
        }

        if( !TaskBarBase::_updateGeometry( action ) )
        {

            _setAction( NoAction );
            setMask( rect() );
            return false;

        } else return true;

    }

    // leave action
    if( action == Leave ) {

        if( !_rootPixmapChanged() )
        {
            setMask( _minimumRect() );
            _setRootPixmapChanged( true );
        }

        if( !TaskBarBase::_updateGeometry( action ) )
        {

            _setAction( NoAction );
            return false;

        } else return true;

    }

    return false;

}

//_________________________________________________________________
void TaskBarMinimum::_preloadSvg( void ) const
{

    Debug::Throw( "TaskBarMinimum::_preloadSvg.\n" );

    Q_ASSERT( _useSvg() );

    // loop over all possible sizes to preload cache
    QSize maximumSize( maximumSize_ );
    QSize minimumSize( minimumRect_.size() );

    Svg::SvgId::List ids;
    const TaskBarAttributes::Layout& layout( attributes().layout() );
    if( layout & ( TaskBarAttributes::Horizontal ) )
    {

        for( int width = minimumSize.width(); width <= maximumSize.width(); width++ )
        { ids << QSize( width, minimumSize.height() ); }

    } else if( layout & (TaskBarAttributes::Horizontal ) ) {

        for( int height = minimumSize.height(); height <= maximumSize.height(); height++ )
        { ids << QSize( minimumSize.width(), height ); }

    } else ids << minimumSize;

    // load sizes
    Svg::SvgEngine::get().preload( ids );

    return;

}
