/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "TaskBarParabolic.h"

#include "CompositeEngine.h"
#include "QtUtil.h"
#include "RoundedRegion.h"
#include "SvgEngine.h"
#include "Util.h"
#include "XmlOptions.h"

#include <QApplication>
#include <QDesktopWidget>
#include <QVector>
#include <algorithm>
#include <cmath>

//____________________________________________________
TaskBarParabolic::TaskBarParabolic( QWidget* parent ):
    TaskBarBase( parent ),
    maxIndex_(0),
    index_(0),
    direction_( TaskButton::NoDirection )
{ Debug::Throw( "TaskBarParabolic::TaskBarParabolic.\n" ); }

//____________________________________________________
void TaskBarParabolic::_updateRootPixmap( void )
{

    _setRootPixmapChanged( false );

    if( Debug::level() >= 1 )
    {
        // solid background
        _setRootPixmap( QPixmap( size() ) );
        _rootPixmap().fill( palette().color( backgroundRole() ) );

    } else if( Transparency::CompositeEngine::get().isAvailable() ) {

        _setRootPixmap( QPixmap( size() ) );
        QPainter painter( &_rootPixmap() );
        painter.setRenderHints(QPainter::SmoothPixmapTransform);
        painter.setCompositionMode(QPainter::CompositionMode_Source );
        painter.fillRect( _rootPixmap().rect(), Qt::transparent);

    } else {

        _setRootPixmap( QPixmap::grabWindow(
            QApplication::desktop()->winId(),
            geometry().x(), geometry().y(),
            geometry().width(), geometry().height() ) );

    }

}

//___________________________________________________
void TaskBarParabolic::_updateBackgroundPixmap( void )
{
    Debug::Throw( "TaskBarParabolic::TaskBarFixed.\n" );

    if( _rootPixmapChanged() )
    {
        // there is no need to update the root pixmap when composite engine is off,
        // unless during resize events, for which the method is called directly
        if( !Transparency::CompositeEngine::get().isAvailable() ) _updateRootPixmap();
        else _setRootPixmapChanged( false );
    }

    TaskBarBase::_updateBackgroundPixmap();

    // need to draw svg on background pixmap if button list is empty
    // due to the way empty taskBars are implemented
    if( _buttons().empty() && _useSvg() && Svg::SvgEngine::get().isValid() )
    {
        QPainter painter( &_backgroundPixmap() );
        painter.setRenderHints(QPainter::SmoothPixmapTransform);
        painter.setCompositionMode(QPainter::CompositionMode_Source );
        painter.drawPixmap( QPoint(0,0),  Svg::SvgEngine::get().get( size() ) );
    }

}

//______________________________________________
void TaskBarParabolic::resizeEvent( QResizeEvent* event )
{
    _updateRootPixmap();

    /*
    need to bypass TaskBarBase::resizeEvent, in order not to have
    the entire background blurred, when enabled
    */
    if( attributes().hasFlag( TaskBarAttributes::RoundCorners ) ) _roundCorners();
    TransparentWidget::resizeEvent( event );
}

//________________________________________
void TaskBarParabolic::enterEvent( QEvent *event )
{

    Debug::Throw( "TaskBarParabolic::enterEvent.\n" );
    if( _mouseButton() == Qt::NoButton ) _setDirection( TaskButton::Bigger );
    TaskBarBase::enterEvent( event );

}

//___________________________________________________
void TaskBarParabolic::leaveEvent( QEvent *event )
{

    Debug::Throw( "TaskBarParabolic::leaveEvent.\n" );
    _setDirection( TaskButton::Smaller );
    TaskBarBase::leaveEvent( event );

}

//___________________________________________________
void TaskBarParabolic::mouseMoveEvent( QMouseEvent *event )
{
    Debug::Throw( "TaskBarParabolic::mouseMoveEvent.\n" );
    TaskBarBase::mouseMoveEvent( event );
    update();
}

//_____________________________________________________________
void TaskBarParabolic::_updateMinimumSize( void )
{

    Debug::Throw( "TaskBarParabolic::_updateMinimumSize.\n" );

    // compute minimum size
    QSize minimumSize;

    // compute visible rect
    const TaskBarAttributes::Layout& layout ( attributes().layout() );
    if( layout & TaskBarAttributes::Vertical )
    {

        for( const auto& button:_buttons() )
        {
            minimumSize.setHeight( minimumSize.height()+button->minimumSize().height()+_spacing() );
            minimumSize.setWidth( qMax<int>( minimumSize.width(), button->minimumSize().width() + _spacing() + _margins().width() ) );
        }

        minimumSize.setHeight( minimumSize.height() + _margins().height() );

    } else if( layout & TaskBarAttributes::Horizontal ) {

        for( const auto& button:_buttons() )
        {
            minimumSize.setWidth( minimumSize.width() + button->minimumSize().width()+_spacing() );
            minimumSize.setHeight( qMax<int>( minimumSize.height(), button->minimumSize().height() + _spacing() + _margins().height() ) );
        }

        minimumSize.setWidth( minimumSize.width() + _margins().width() );

    } else {

        int columnIndex(0);
        int lineIndex(0);

        // compute column width/line height
        QVector<int> columnWidth( _numColumns(), 0 );
        QVector<int> lineHeight( _numLines(), 0 );
        for( const auto& button:_buttons() )
        {

            lineHeight[lineIndex] = qMax( lineHeight[lineIndex], button->minimumSize().height() + _spacing() );
            columnWidth[columnIndex] = qMax( columnWidth[columnIndex], button->minimumSize().width() + _spacing() );
            columnIndex++;
            if( columnIndex >= _numColumns() ) {
                columnIndex = 0;
                lineIndex++;
            }

        }

        int lineWidth(0);
        columnIndex = 0;
        lineIndex = 0;
        for( const auto& button:_buttons() )
        {

            Q_UNUSED( button );
            lineWidth += columnWidth[columnIndex];
            columnIndex++;

            // change column
            if( columnIndex >= _numColumns() )
            {
                minimumSize.setWidth( qMax( minimumSize.width(), lineWidth ) );
                minimumSize.setHeight( minimumSize.height()+lineHeight[lineIndex] );

                lineWidth = 0;
                columnIndex = 0;
                lineIndex++;

            }

        }

        minimumSize.setHeight( minimumSize.height() + lineHeight.back() + _margins().height() );
        minimumSize.setWidth( minimumSize.width() + _margins().width() );

    }

    // update
    _setMinimumSize( minimumSize );
    return;

}

//_____________________________________________________________
void TaskBarParabolic::_updateMinimumRectPosition( void )
{

    Debug::Throw( "TaskBarParabolic::_updateMinimumRectPosition.\n" );

    QRect minimumRect( _minimumRect() );
    QPoint offset;
    const TaskBarAttributes::Layout& layout ( attributes().layout() );
    if( layout & TaskBarAttributes::Vertical ) {

        offset.setY( ( _maximumSize().height() - minimumRect.height() )/2 );
        if( layout & TaskBarAttributes::BottomRight ) offset.setX( _maximumSize().width() - minimumRect.width() );
        else if( layout & TaskBarAttributes::Center ) offset.setX( ( _maximumSize().width() - minimumRect.width() )/2 );

    } else if( layout & TaskBarAttributes::Horizontal ) {

        offset.setX( ( _maximumSize().width() - minimumRect.width() )/2 );
        if( layout & TaskBarAttributes::BottomRight ) offset.setY( _maximumSize().height() - minimumRect.height() );
        else if( layout & TaskBarAttributes::Center ) offset.setY( ( _maximumSize().height() - minimumRect.height() )/2 );

    } else  {

        offset.setX( ( _maximumSize().width() - minimumRect.width() )/2 );
        offset.setY( ( _maximumSize().height() - minimumRect.height() )/2 );

    }

    minimumRect.moveTo( offset );
    _setMinimumRect( minimumRect );

}

//_____________________________________________________________
void TaskBarParabolic::_updateVisibleRect( void )
{

    Debug::Throw( "TaskBarParabolic::_updateVisibleRect.\n" );

    QRect visibleRect;
    const TaskBarAttributes::Layout& layout ( attributes().layout() );
    if( layout & TaskBarAttributes::Vertical )
    {

        // width is from minimum size
        visibleRect.setWidth( _minimumSize().width() );

        // when compositing is enabled one can shrink the visible rect
        // it does not work otherwise, due to 'echoes' in the root pixmap
        if( Transparency::CompositeEngine::get().isAvailable() && _buttons().size() > 1 )
        {

            // get bounding height and y offset
            bool first( true );
            for( const auto& button:_buttons() )
            {
                visibleRect.setTop( first ? button->rect().top() : qMin( visibleRect.top(), button->rect().top() ) );
                visibleRect.setBottom( qMax( visibleRect.bottom(), button->rect().bottom() ) );
                first = false;
            }

            visibleRect.setTop( visibleRect.top() - _margins().top() );
            visibleRect.setBottom( visibleRect.bottom() + _margins().bottom() );

        } else {

            visibleRect.setHeight( _maximumSize( _index() ).height() );
            visibleRect.moveTop( (height() - _maximumSize( _index() ).height())/2 );

        }

        if( layout & TaskBarAttributes::BottomRight ) visibleRect.moveLeft(  _maximumSize().width() - visibleRect.width() );
        else if( layout & TaskBarAttributes::Center ) visibleRect.moveLeft( ( _maximumSize().width() - visibleRect.width() )/2 );

    } else if( layout & TaskBarAttributes::Horizontal ) {

        // height is from minimum size
        visibleRect.setHeight( _minimumSize().height() );

        // when compositing is enabled one can shrink the visible rect
        // it does not work otherwise, due to 'echoes' in the root pixmap
        if( Transparency::CompositeEngine::get().isAvailable() && _buttons().size() > 1  )
        {

            // get bounding height and y offset
            bool first( true );
            for( const auto& button:_buttons() )
            {
                visibleRect.setLeft( first ? button->rect().left() : qMin( visibleRect.left(), button->rect().left() ) );
                visibleRect.setRight( qMax( visibleRect.right(), button->rect().right() ) );
                first = false;
            }

            visibleRect.setLeft( visibleRect.left() - _margins().left() );
            visibleRect.setRight( visibleRect.right() + _margins().right() );

        } else {

            visibleRect.setWidth( _maximumSize( _index() ).width() );
            visibleRect.moveLeft( (width() - _maximumSize( _index() ).width())/2 );

        }

        if( layout & TaskBarAttributes::BottomRight ) visibleRect.moveTop(  _maximumSize().height() - visibleRect.height() );
        else if( layout & TaskBarAttributes::Center ) visibleRect.moveTop( ( _maximumSize().height() - visibleRect.height() )/2 );

    } else {

        visibleRect.setSize( _minimumSize() );
        visibleRect.moveTo( ( _maximumSize().width() - visibleRect.width() )/2, ( _maximumSize().height() - visibleRect.height() )/2 );

    }

    // add offset (for auto-hide)
    visibleRect.translate( _offset() );

    // in case of autoHide, one make sure there is at least
    // a thin band drawn at the bottom, that match the hidden geometry
    if( attributes().hasFlag( TaskBarAttributes::AutoHide) )
    {

        const TaskBarAttributes::Layout& layout ( attributes().layout() );
        if( layout & TaskBarAttributes::Horizontal )
        {

            if( layout & TaskBarAttributes::TopLeft )  visibleRect.setBottom( qMax(  visibleRect.bottom(), 3 ) );
            else visibleRect.setTop( qMin( visibleRect.top(), height()-3 ) );

        } else {

            if( layout & TaskBarAttributes::TopLeft ) visibleRect.setRight( qMax( visibleRect.right(), 3 ) );
            else visibleRect.setLeft( qMin( visibleRect.left(), width()-3 ) );

        }

    }

    // update blur areas if visibleRect is changed
    if( _setVisibleRect( visibleRect ) && Transparency::CompositeEngine::get().isAvailable() )
    { _updateBlurRegion( _outerPadding().adjustedRect( visibleRect ).translated( mapTo( window(), QPoint(0,0) ) ) ); }

    return;

}


//_____________________________________________________________
bool TaskBarParabolic::_updateGeometry( Action action )
{

    Debug::Throw() << "TaskBarParabolic::_updateGeometry - action: " << action << endl;

    if( action == NoAction ) return false;

    // minimum action
    if( action == Minimize )
    {

        setMask( _minimumRect() );
        if( !Transparency::CompositeEngine::get().isAvailable() ) _setRootPixmapChanged( true );
        _setAction( NoAction );
        return false;

    }

    // enter action
    if( action == Enter ) {

        if( _rootPixmapChanged() )
        {
            // there is no need to update the root pixmap when composite engine is off,
            // unless during resize events, for which the method is called directly
            if( !Transparency::CompositeEngine::get().isAvailable() ) _updateRootPixmap();
            else _setRootPixmapChanged( false );
        }


        if( !TaskBarBase::_updateGeometry( action ) )
        {

            _setAction( NoAction );
            setMask( rect() );
            return false;

        } else return true;

    }

    // leave action
    if( action == Leave )
    {

        if( !_rootPixmapChanged() )
        {
            setMask( _minimumRect() );
            _setRootPixmapChanged( true );
        }

        if( !TaskBarBase::_updateGeometry( action ) )
        {

            _setAction( NoAction );
            return false;

        } else return true;

    }

    return false;

}

//_________________________________________________________________
void TaskBarParabolic::_updateMaximumSize( void )
{

    Debug::Throw( "TaskBarParabolic::_updateMaximumSize.\n" );

    // maxumum size and index
    _setMaximumIndex( (*std::min_element( _buttons().begin(), _buttons().end(), TaskButton::MaxIndexFTor() ))->maximumIndex() );
    _maximumButtonSize().setWidth( (*std::max_element( _buttons().begin(), _buttons().end(), TaskButton::MaxWidthFTor() ))->maximumSize().width() + _spacing() );
    _maximumButtonSize().setHeight( (*std::max_element( _buttons().begin(), _buttons().end(), TaskButton::MaxHeightFTor() ))->maximumSize().height() + _spacing() );

    // reset button indices
    _setIndex(0);
    _placeButtons();

    // reset mouse offset
    _setMouseOffset( QPoint() );
    _clearMaximumSizes();
    _clearMidPositions();
    for( int index = 0; index < _maximumIndex(); index++ )
    {

        QSize maximumSize;

        // retrieve taskBar layout
        const TaskBarAttributes::Layout& layout( attributes().layout() );
        if( layout & TaskBarAttributes::Horizontal ) maximumSize.setHeight( _maximumButtonSize().height() );
        else if( layout & TaskBarAttributes::Vertical ) maximumSize.setWidth( _maximumButtonSize().width() );

        // store the mid position
        QPoint mid_position;
        int counts( 0 );

        // set index to max value
        _setIndex( index );

        // increment mouse position over the entire taskBar to derive the maximum size
        for( const auto& button:_buttons() )
        {

            // set mouse position
            // the use of 'center' position when there is only one button is to fix display
            // why it is not usable for button size larger than 1 is unclear.
            // whether the second works for other orientations is also unclear
            // -> to be fixed
            if( _buttons().size() == 1 ) _setMousePosition( button->rect().center() );
            else _setMousePosition( button->rect().topLeft() );

            // place buttons
            _placeButtons();

            if( layout & TaskBarAttributes::Horizontal )
            {

                // get bounding rect
                int boundingSize(0);
                for( const auto& button:_buttons() )
                { boundingSize = qMax( boundingSize, button->rect().right() ); }

                // update maximum size
                if( boundingSize > maximumSize.width() )
                {
                    maximumSize.setWidth( boundingSize );
                    mid_position = _mousePosition();
                    counts = 1;

                } else if( boundingSize == maximumSize.width() ) {

                    mid_position += _mousePosition();
                    counts++;

                }

            } else if( layout & TaskBarAttributes::Vertical ) {

                // get bounding rect
                int boundingSize(0);
                for( const auto& button:_buttons() )
                { boundingSize = qMax( boundingSize, button->rect().bottom() ); }

                // update maximum size
                if( boundingSize > maximumSize.height() )
                {
                    maximumSize.setHeight( boundingSize );
                    mid_position = _mousePosition();
                    counts = 1;

                } else if( boundingSize == maximumSize.height() ) {

                    mid_position += _mousePosition();
                    counts++;

                }

            } else {

                // get bounding rect
                for( const auto& button:_buttons() )
                {
                    maximumSize.setWidth( qMax( maximumSize.width(), button->rect().left() ) );
                    maximumSize.setHeight( qMax( maximumSize.height(), button->rect().bottom() ) );
                }

                // set mouse position
                _setMousePosition( button->rect().center() );
                _placeButtons();

                // get bounding rect
                for( const auto& button:_buttons() )
                {
                    maximumSize.setWidth( qMax( maximumSize.width(), button->rect().left() ) );
                    maximumSize.setHeight( qMax( maximumSize.height(), button->rect().bottom() ) );
                }

                // set mouse position
                _setMousePosition( button->rect().bottomRight() );
                _placeButtons();

                // get bounding rect
                for( const auto& button:_buttons() )
                {
                    maximumSize.setWidth( qMax( maximumSize.width(), button->rect().left() ) );
                    maximumSize.setHeight( qMax( maximumSize.height(), button->rect().bottom() ) );
                }

            }

        }

        // store maximum size and mid position
        if( counts ) _addMidPosition( mid_position/counts + _margins().topLeft() );
        _addMaximumSize( QSize( maximumSize.width() + _margins().width(), maximumSize.height() + _margins().height() ) );

    }

    // reset index
    _setIndex(0);
    return;

}


//______________________________________________________
void TaskBarParabolic::_resize( void )
{

    Debug::Throw( "TaskBarParabolic::_resize.\n" );

    // if no tasks, set empty message and exit
    if( _buttons().empty() ) {
        resize( emptySize() );
        return;
    }

    // dimensions
    _setDimensions();
    _updateMinimumSize();
    _updateMaximumSize();
    _updateMinimumRectPosition();

    // svg
    if( _useSvg() && Svg::SvgEngine::get().isValid() ) _preloadSvg();

    // make sure taskBar is in minimum state
    // resize
    _setAction( Minimize );

    resize( _maximumSize() );

}

//_________________________________________________________________
void TaskBarParabolic::_placeButtons( void )
{

    // compute TaskButtons position
    int x(0);
    int y(0);

    const TaskBarAttributes::Layout& layout( attributes().layout() );
    if( layout & (TaskBarAttributes::Vertical | TaskBarAttributes::Horizontal ) )
    {

        for( const auto& button:_buttons() )
        {

            // place button at x and y
            button->rect().moveTo( x, y );

            // calculate index and assign
            button->setIndex( _index( button ) );
            button->rect().setSize( QSize( button->currentSize().width() + _spacing(), button->currentSize().height() + _spacing() ) );

            // update x and y position
            if( layout & TaskBarAttributes::Vertical ) y+= button->rect().height();
            else if( layout & TaskBarAttributes::Horizontal ) x += button->rect().width();

        }

    } else {

        int columnIndex(0);
        int lineIndex(0);

        // compute column width/line height
        QVector<int> columnWidth( _numColumns(), 0 );
        QVector<int> lineHeight( _numLines(), 0 );
        for( const auto& button:_buttons() )
        {

            button->setIndex( _index( button ) );
            lineHeight[lineIndex] = qMax( lineHeight[lineIndex], button->currentSize().height() + _spacing() );
            columnWidth[columnIndex] = qMax( columnWidth[columnIndex], button->currentSize().width() + _spacing() );
            columnIndex++;
            if( columnIndex >= _numColumns() ) {
                columnIndex = 0;
                lineIndex++;
            }

        }

        // place buttons
        columnIndex = 0;
        lineIndex = 0;
        for( const auto& button:_buttons() )
        {
            button->setRect( QRect( QPoint( x, y ), QSize( columnWidth[columnIndex], lineHeight[lineIndex] ) ) );
            x+=columnWidth[columnIndex];
            columnIndex++;
            if( columnIndex >= _numColumns() )
            {
                x = 0;
                columnIndex = 0;
                y+= lineHeight[lineIndex];
                lineIndex++;
            }
        }

    }

}


//______________________________________________________
void TaskBarParabolic::_addOffsets( void )
{

    // get bounding rect for current button sizes
    QRect boundingRect;
    for( const auto& button:_buttons() )
    {
        boundingRect.setWidth( qMax( boundingRect.width(), button->rect().right() ) );
        boundingRect.setHeight( qMax( boundingRect.height(), button->rect().bottom() ) );
    }

    // compute offsets
    const TaskBarAttributes::Layout &layout( attributes().layout() );

    QPoint offset( _margins().topLeft() );
    if( (layout & TaskBarAttributes::Horizontal) )
    {

        if( _mousePosition().x() > _midPosition( _index() ).x() ) offset.setX( (width() - _maximumSize( _index() ).width())/2 + _margins().left() );
        else offset.setX( ( width() + _maximumSize( _index() ).width())/2 - boundingRect.width() - _margins().right() );

    } else if( (layout & TaskBarAttributes::Vertical) ) {

        if( _mousePosition().y() > _midPosition( _index() ).y() ) offset.setY( (height() - _maximumSize( _index() ).height())/2 + _margins().top() );
        else offset.setY( ( height() + _maximumSize( _index() ).height())/2 - boundingRect.height() - _margins().bottom() );

    } else {

        offset = QPoint( (size().width() - boundingRect.width())/2, (size().height() - boundingRect.height())/2 );

    }

    _setMouseOffset( offset );

    // add offsets
    for( const auto& button:_buttons() )
    {

        QPoint buttonOffset( offset );
        if( layout & TaskBarAttributes::Vertical )
        {

            if( layout & TaskBarAttributes::TopLeft ) buttonOffset.setX( _margins().left() );
            else if( layout & TaskBarAttributes::BottomRight ) buttonOffset.setX( width() - button->rect().width() - _margins().right() );
            else if( layout & TaskBarAttributes::Center ) buttonOffset.setX( (width() - button->rect().width())/2 );

        } else if( layout & TaskBarAttributes::Horizontal ) {

            if( layout & TaskBarAttributes::TopLeft ) buttonOffset.setY( _margins().top() );
            else if( layout & TaskBarAttributes::BottomRight ) buttonOffset.setY( height() - button->rect().height() - _margins().bottom() );
            else if( layout & TaskBarAttributes::Center ) buttonOffset.setY( (height() - button->rect().height())/2 );

        }

        button->setRect( button->rect().translated( buttonOffset ) );

    }

}

//___________________________________________________
bool TaskBarParabolic::_drawButtons( QPaintDevice& device, QRect rect )
{

    // if no tasks, do nothing
    if( _buttons().empty() ) return false;

    // place the buttons
    _placeButtons();
    _addOffsets();

    // redraw
    QPainter painter( &device );
    painter.setClipRect( rect );

    if( Transparency::CompositeEngine::get().isAvailable() )
    {
        painter.setRenderHints(QPainter::SmoothPixmapTransform);
        painter.setCompositionMode(QPainter::CompositionMode_Source );
    }

    // draw root pixmap into entire window
    painter.drawPixmap( 0, 0, _rootPixmap() );

    // draw background pixmap clipped to visibleRect
    _updateVisibleRect();
    _drawVisibleRect( painter );

    // draw buttons, translated by offset if needed
    painter.setCompositionMode(QPainter::CompositionMode_SourceOver );
    painter.translate( _offset() );
    for( const auto& button:_buttons() )
    { button->paint( painter ); }

    painter.end();

    // update index
    int increment = attributes().hasFlag( TaskBarAttributes::ZoomEnabled ) ? 1:_maximumIndex()-1;
    if( _direction() == TaskButton::Bigger && _index() < _maximumIndex() -1 ) { if( _action() != Enter ) _setIndex( _index()+increment ); }
    else if( _direction() == TaskButton::Smaller && _index() > 0 ) _setIndex( _index()-increment );
    else _setDirection( TaskButton::NoDirection );

    bool out( false );
    for( const auto& button:_buttons() )
    {

        if( _action() != Enter )
        {
            button->updateHighlightDirection();
            out |= button->highlightDirection() != TaskButton::NoHighlight;
        }

    }

    return out || _direction() != TaskButton::NoDirection;

}

//___________________________________________________
void TaskBarParabolic::_drawVisibleRect( QPainter& painter )
{
    painter.save();

    if( _useSvg() && Svg::SvgEngine::get().isValid() )
    {

        painter.translate( _visibleRect().topLeft() );
        painter.drawPixmap( QPoint(0,0), Svg::SvgEngine::get().get( _visibleRect().size() ) );

    } else {

        if( attributes().hasFlag( TaskBarAttributes::RoundCorners )  ) painter.setClipRegion( RoundedRegion( _visibleRect(), RoundedRegion::All ) );
        else painter.setClipRect( _visibleRect() );
        painter.drawPixmap( 0, 0, _backgroundPixmap() );

    }
    painter.restore();
}

//___________________________________________________
void TaskBarParabolic::_setEnableMove( bool value )
{
    Debug::Throw( "TaskBar::_setEnableMove.\n" );
    if( _moveEnabled() == value ) return;
    _setDirection( TaskButton::Smaller );
    TaskBarBase::_setEnableMove( value );
}

//_________________________________________________________________
int TaskBarParabolic::_index( const TaskButton::Pointer& button )
{

    if( !_index() ) return 0;

    double distance(0);
    const TaskBarAttributes::Layout& layout( attributes().layout() );
    if( layout & TaskBarAttributes::Horizontal )
    {

        distance = fabs(double( _mousePosition().x() - _mouseOffset().x() ) - double( button->rect().left() + button->size(_index()).width()/2 ) );
        distance /= _buttons().size()*( _maximumButtonSize().width() + _spacing() );

    } else if( layout & TaskBarAttributes::Vertical ) {

        distance = fabs( double( _mousePosition().y() - _mouseOffset().y() ) - double( button->rect().top() + button->size(_index()).height()/2 ) );
        distance /= _buttons().size()*( _maximumButtonSize().height() + _spacing() );

    } else {

        distance = 0;
        if( _numColumns() > 1 ) {
            distance += fabs(double( _mousePosition().x() ) - double( button->rect().left() + button->size(_index()).width()/2 ) )/( _maximumButtonSize().width() + _spacing() )/_numColumns();
        }

        if( _numLines() > 1 ) {
            distance += fabs(double( _mousePosition().y() ) - double( button->rect().top() + button->size(_index()).height()/2 ) )/( _maximumButtonSize().height() + _spacing() )/_numLines();
        }

    }

    return _index( distance );

}

//_________________________________________________________________
int TaskBarParabolic::_index( const double& distance )
{

    if( !_index() ) return 0;
    double out( _index()*(1-Util::sqr(attributes().scale()*distance) ) );
    if( out < 0 ) return 0;
    if( out > _maximumIndex()-1 ) return _maximumIndex()-1;
    return (unsigned int) out;

}

//_________________________________________________________________
void TaskBarParabolic::_preloadSvg( void ) const
{

    Debug::Throw( "TaskBarParabolic::_preloadSvg.\n" );
    Q_ASSERT( _useSvg() && !maximumSizes_.empty() );

    // loop over all possible sizes to preload cache
    QSize maximumSize( maximumSizes_.back() );
    QSize minimumSize( minimumRect_.size() );

    const TaskBarAttributes::Layout& layout( attributes().layout() );
    Svg::SvgId::List ids;
    if( layout & ( TaskBarAttributes::Horizontal ) )
    {

        Debug::Throw() << "TaskBarParabolic::_preloadSvg - (" << minimumSize.width() << "," << maximumSize.width() << ")" << endl;

        for( int width = minimumSize.width(); width <= maximumSize.width(); width++ )
        { ids << QSize( width, minimumSize.height() ); }

    } else if( layout & (TaskBarAttributes::Vertical ) ) {

        for( int height = minimumSize.height(); height <= maximumSize.height(); height++ )
        { ids << QSize( minimumSize.width(), height ); }

    } else ids << minimumSize;

    Svg::SvgEngine::get().preload( ids );

    return;
}
