#ifndef TaskBarParabolic_h
#define TaskBarParabolic_h


/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "TaskBarBase.h"

#include <QSize>
#include <QList>

//! TaskBar main widget
class TaskBarParabolic: public TaskBarBase
{

    public:

    //! constructor
    TaskBarParabolic( QWidget* = 0 );

    protected Q_SLOTS:

    //! root pixmap
    virtual void _updateRootPixmap( void );

    //! background background
    virtual void _updateBackgroundPixmap( void );

    protected:

    //!@name event handlers
    //@{

    //! resize event method
    virtual void resizeEvent( QResizeEvent* );

    //! pointer leaves window
    virtual void enterEvent( QEvent *event );

    //! pointer leaves window
    virtual void leaveEvent( QEvent *event );

    //! mouse move event
    virtual void mouseMoveEvent( QMouseEvent *event );

    //@}

    //! resize the taskBar so that is is able to contain all buttons
    virtual void _resize( void );

    //! place buttons in the taskBar
    void _placeButtons( void );

    //! adds offset to the buttons
    virtual void _addOffsets( void );

    //! draw buttons
    /*!
    if force is true, only buttons which are zooming are drawn
    returns true if at least one if zooming
    */
    virtual bool _drawButtons( QPaintDevice&, QRect );

    //! draw visible rect
    virtual void _drawVisibleRect( QPainter& );

    //! update minimum size
    virtual void _updateMinimumSize( void );

    //! update minimum rect
    virtual void _updateMinimumRectPosition( void );

    //! minimum size
    virtual QSize _minimumSize() const
    { return minimumRect_.size(); }

    //! minimum size
    virtual void _setMinimumSize( const QSize& size )
    { minimumRect_.setSize( size ); }

    //! minimum size
    virtual QRect _minimumRect() const
    { return minimumRect_; }

    //! minimum size
    virtual void _setMinimumRect( const QRect& rect )
    { minimumRect_ = rect; }

    //! update maximum size
    void _updateMaximumSize( void );

    //! maximum size
    virtual const QSize& _maximumSize() const
    { return maximumSizes_.back(); }

    //! maximum size
    virtual const QSize& _maximumSize( const int& index ) const
    { return maximumSizes_[index]; }

    //! maximum size
    virtual void _setMaximumSize( const QSize& )
    {}

    //! mid position
    virtual void _clearMaximumSizes( void )
    { maximumSizes_.clear(); }

    //! maximum size
    virtual void _addMaximumSize( const QSize& size )
    { maximumSizes_ << size; }

    //! mid position, used to switch from one side of the bar to the other
    virtual const QPoint& _midPosition( const int& index ) const
    { return midPositions_[index]; }

    //! mid position
    virtual void _clearMidPositions( void )
    { midPositions_.clear(); }

    //! mid position, used to switch from one side of the bar to the other
    virtual void _addMidPosition( const QPoint& value )
    { midPositions_ << value; }

    //! mouse offset to get the proper size when one is on the "wrong" side of the bar
    virtual const QPoint& _mouseOffset() const
    { return mouseOffset_; }

    //! mouse offset to get the proper size when one is on the "wrong" side of the bar
    virtual void _setMouseOffset( const QPoint& value )
    { mouseOffset_ = value; }

    //! update visible rect
    virtual void _updateVisibleRect( void );

    //! update geometry (mask) for enter/leave event
    virtual bool _updateGeometry( Action action );

    //! max index
    int _maximumIndex( void ) const
    { return maxIndex_; }

    //! max index
    void _setMaximumIndex( const unsigned int& index )
    { maxIndex_ = index; }

    //! max index
    QSize& _maximumButtonSize( void )
    { return maxButtonSize_; }

    //! max index
    void _setMaximumButtonSize( const QSize& size )
    { maxButtonSize_ = size; }

    //! current index
    int _index( void ) const
    { return index_; }

    //! current index
    void _setIndex( const unsigned int& index )
    { index_ = index; }

    //! direction
    const TaskButton::Direction& _direction( void ) const
    { return direction_; }

    //! zoom direction
    void _setDirection( const TaskButton::Direction& direction )
    { direction_ = direction; }

    //! enable move
    virtual void _setEnableMove( bool value = true );

    //! visible rect
    virtual const QRect& _visibleRect( void ) const
    { return visible_rect_; }

    //! visible rect
    virtual QRect& _visibleRect( void )
    { return visible_rect_; }

    //! visible rect
    virtual bool _setVisibleRect( const QRect& rect )
    {
        if( visible_rect_ == rect ) return false;
        visible_rect_ = rect;
        return true;
    }

    //! root pixmap
    virtual QPixmap _rootPixmap( void ) const
    { return rootPixmap_; }

    //! root pixmap
    virtual QPixmap& _rootPixmap( void )
    { return rootPixmap_; }

    //! root pixmap
    virtual void _setRootPixmap( QPixmap pixmap )
    { rootPixmap_ = pixmap; }

    //! root pixmap needs update
    virtual bool _rootPixmapChanged( void ) const
    { return rootPixmapChanged_; }

    //! root pixmap needs update
    virtual void _setRootPixmapChanged( bool value )
    { rootPixmapChanged_ = value; }

    private:

    //! get button size index
    int _index( const TaskButton::Pointer& button );

    //! get button size index
    int _index( const double& distance );

    //! preload svg
    void _preloadSvg( void ) const;

    //! taskBar minimum size
    QRect minimumRect_;

    //! taskBar maximum size
    QList<QSize> maximumSizes_;

    //! mid position (that makes the bar switch from one anchor to the other)
    QList<QPoint> midPositions_;

    //! mouse offset
    QPoint mouseOffset_;

    //! max index
    int maxIndex_;

    //! max size
    QSize maxButtonSize_;

    //! current index
    unsigned int index_;

    //! zoom direction
    TaskButton::Direction direction_;

    //! taskBar visible rect
    /*! this represents the rectangle that is shaded and on top of which the buttons are drawn */
    QRect visible_rect_;

    //! root pixmap
    QPixmap rootPixmap_;

    //! root pixmap needs update
    bool rootPixmapChanged_;

};

#endif
