#ifndef TaskButton_h
#define TaskButton_h

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Counter.h"
#include "PixmapList.h"
#include "TaskAttributes.h"

#include <QColor>
#include <QPainter>
#include <QPixmap>
#include <QTimer>
#include <QList>

//* taskBar button
class TaskButton: public QObject, public Counter
{

    public:

    //* shortcut to list of buttons
    using Pointer = TaskButton*;
    using List = QList<Pointer>;

    //* null pointer
    static Pointer null( void )
    { return nullptr; }

    //* constructor
    TaskButton( const TaskBarAttributes& , const TaskAttributes& );

    //* equal to operator
    bool operator == ( const TaskButton& button ) const
    { return attributes().id() == button.attributes().id(); }

    //*@name accessors
    //@{

    //* true if button is valid
    bool isValid( void ) const
    { return valid_; }

    //* Rect
    const QRect& rect( void ) const
    { return rect_; }

    //* size of pixmap for index id
    QSize size( int ) const;

    //* maximum size
    QSize minimumSize( void ) const
    { return size(0); }

    //* current size
    QSize currentSize( void ) const
    { return size( index_ ); }

    //* maximum size
    QSize maximumSize( void ) const
    { return size( maximumIndex()-1 ); }

    //* text size
    const QSize& textSize( void ) const
    { return textSize_; }

    //* attributes
    const TaskAttributes& attributes( void ) const
    { return attributes_; }

    //* Need resize
    bool needResize( void ) const
    { return needResize_; }

    //* highlight
    bool hasHighlightedPixmaps( void ) const
    { return pixmaps_.maxHighlightIndex() > 1; }

    //* highlight direction
    enum HighlightDirection
    {
        Lighter,
        Darker,
        NoHighlight
    };

    //* update highlight direction
    HighlightDirection highlightDirection( void ) const
    { return highlightDirection_; }

    //* highlight index
    int highlightIndex( void ) const
    { return trueHighlightIndex_; }

    //* zoom direction
    enum Direction
    {
        Bigger,
        Smaller,
        NoDirection
    };

    //* zoom direction
    const Direction& direction( void ) const
    { return direction_; }

    //* max index (biggest zoom)
    int maximumIndex( void ) const
    { return pixmaps_.maxSizeIndex(); }

    //* current index
    int index( void ) const
    { return index_; }

    //@}

    //*@name modifiers
    //@{

    //* Rect
    void setRect( const QRect& rect )
    { rect_ = rect; }

    //* Rect
    void setRect( const QPoint& point, const QSize& size )
    { rect_ = QRect( point, size ); }

    //* Rect
    QRect& rect( void )
    { return rect_; }

    //* attributes
    void setAttributes( const TaskBarAttributes& barAttributes, const TaskAttributes& attributes );

    //* attributes
    TaskAttributes& attributes( void )
    { return attributes_; }

    //@name highlight
    //@{


    //* highlight
    void highlight( bool value )
    {
        highlight_ = value;
        if( highlight_ ) highlightDirection_ = Lighter;
    }

    //* update highlight direction
    HighlightDirection updateHighlightDirection( void );

    //@name zoom
    //@{

    //* zoom direction
    void setDirection( const Direction& );

    //* current index
    void setIndex( int index )
    {
        Q_ASSERT( index < maximumIndex() );
        index_ = index;
    }

    //* update direction
    /** direction is changed when button is at maximum size */
    Direction updateDirection( void );

    //@}

    //* erase
    /** fill the button rect() geometry with the background pixmap */
    void erase( QPainter& painter, QPixmap background ) const
    { painter.drawPixmap( rect_.topLeft(), background, rect_ ); }

    //* paint
    /*
    redraw itself at proper position in painter
    \param painter, the destination painter
    the current index is used to choose pixmap.
    It is updated according to the button zoom direction.
    */
    void paint( QPainter& painter ) const;

    //@}

    //* used to retrieve a button from point position
    class ContainsFTor
    {

        public:

        //* constructor
        ContainsFTor( const QPoint& point ):
            point_( point )
        {}

        //* predicate
        bool operator() (const TaskButton& button) const
        { return button.rect().contains( point_ ); }

        //* predicate
        bool operator() (const Pointer& button) const
        { return operator()( *button ); }

        private:

        //* prediction point
        const QPoint& point_;

    };

    //* used to retrieve a button from attributes
    class SameAttributesFTor
    {

        public:

        //* constructor
        SameAttributesFTor( const TaskAttributes& attributes ):
            attributes_( attributes )
        {}

        //* predicate
        bool operator() (const TaskButton& button ) const
        { return button.attributes().id() == attributes_.id(); }

        //* predicate
        bool operator() (const Pointer& button) const
        { return operator()( *button ); }

        private:

        //* predicted attributes
        const TaskAttributes& attributes_;

    };

    //* used to get button of maximum width
    class MaxWidthFTor
    {

        public:

        //* predicate
        bool operator() (const TaskButton& first, const TaskButton& second ) const
        { return first.maximumSize().width() < second.maximumSize().width(); }

        //* predicate
        bool operator() (const Pointer& first, const Pointer& second ) const
        { return operator()( *first, *second ); }

    };

    //* used to get button of maximum height
    class MaxHeightFTor
    {

        public:

        //* predicate
        bool operator() (const TaskButton& first, const TaskButton& second ) const
        { return first.maximumSize().height() < second.maximumSize().height(); }

        //* predicate
        bool operator() (const Pointer& first, const Pointer& second ) const
        { return operator()( *first, *second ); }

    };

    //* used to get button of minimum width
    class MinWidthFTor
    {

        public:

        //* predicate
        bool operator() (const TaskButton& first, const TaskButton& second ) const
        { return first.minimumSize().width() < second.minimumSize().width(); }

        //* predicate
        bool operator() (const Pointer& first, const Pointer& second ) const
        { return operator()( *first, *second ); }

    };

    //* used to get button of minimum height
    class MinHeightFTor
    {

        public:

        //* predicate
        bool operator() (const TaskButton& first, const TaskButton& second ) const
        { return first.minimumSize().height() < second.minimumSize().height(); }

        //* predicate
        bool operator() (const Pointer& first, const Pointer& second ) const
        { return operator()( *first, *second ); }

    };

    //* used to get the button with the maximum number of mixmaps
    class MaxIndexFTor
    {
        public:

        //* predicate
        bool operator() (const TaskButton& first, const TaskButton& second ) const
        {
            if( first.attributes().hasFlag( TaskAttributes::IsSeparator ) ) return false;
            if( second.attributes().hasFlag( TaskAttributes::IsSeparator ) ) return true;
            return first.maximumIndex() < second.maximumIndex();
        }

        //* predicate
        bool operator() (const Pointer& first, const Pointer& second ) const
        { return operator()( *first, *second ); }

    };

    private:

    //* need resize
    void _setNeedResize( bool value )
    { needResize_ = value; }

    //* validity
    void _setIsValid( bool validity )
    { valid_ = validity; }

    //* draw task pixmap in painter
    void _drawPixmap( QPainter& painter, int index ) const;

    //* draw task name in painter
    void _drawText( QPainter& painter, int index ) const;

    //* update highlight index
    void _updateTrueHighlightIndex( void );

    //* associated task
    TaskAttributes attributes_;

    //*@name geometry
    //@{

    //* pixmap position in painter
    QRect rect_;

    //* true if new maximumSize is bigger after SetAttributes
    bool needResize_ = false;

    //@}

    //*@name button name
    //*@{

    //* true if name is to be drawn
    bool showName_ = false;

    //* text size
    QSize textSize_;

    //* shadow offset
    int shadowOffset_ = 0;

    //* foreground color
    QColor foregroundColor_;

    //* shadow color
    QColor shadowColor_;

    //@}

    //* vector of scaled pixmap from small to high
    PixmapList pixmaps_;

    //* true if highlighted
    bool highlight_  = false;

    //* maximum highlightIndex_
    int maximumHighlightIndex_ = 0;

    //* highlight index
    double highlightIndex_ = 0;

    //* highlight index
    int trueHighlightIndex_ = 0;

    //* highlight direction
    HighlightDirection highlightDirection_ = NoHighlight;

    //* current position in pixmap vector
    int index_ = 0;

    //* zoom direction
    Direction direction_ = NoDirection;

    //* validity
    bool valid_ = false;

};

#endif
