/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "TaskDialog.h"

#include "BrowseIconButton.h"
#include "BrowsedLineEditor.h"
#include "Debug.h"
#include "GridLayout.h"
#include "LineEditor.h"
#include "TextEditor.h"

#include <QLayout>
#include <QLabel>

//______________________________________________
TaskDialog::TaskDialog( QWidget* parent ):
    CustomDialog( parent )
{

    Debug::Throw( "TaskDialog::TaskDialog.\n" );
    setOptionName( "TASK_DIALOG" );
    setWindowTitle( tr( "Edit Button - Taskbar" ) );

    QHBoxLayout* hLayout = new QHBoxLayout();
    hLayout->setMargin(0);
    hLayout->setSpacing(5);
    mainLayout().addLayout( hLayout, 0 );

    QVBoxLayout* vLayout;
    hLayout->addLayout( vLayout = new QVBoxLayout() );
    vLayout->setMargin(10);

    vLayout->addWidget( iconEdit_ = new BrowseIconButton( this ) );
    iconEdit_->setFile( TaskAttributes().defaultIconFilename(), false );
    vLayout->addStretch(1);

    hLayout->addLayout( vLayout = new QVBoxLayout() );
    vLayout->setMargin(0);
    vLayout->setSpacing(5);

    QLabel* label;
    vLayout->addWidget( label = new QLabel( tr( "Name:" ), this ) );
    vLayout->addWidget( nameEditor_ = new LineEditor( this ) );
    nameEditor_->setToolTip(
        tr( "Button name.\n"
        "It is displayed in tooltip and"
        "possibly in the task bar" ) );
    label->setBuddy( nameEditor_ );

    vLayout->addWidget( showNameCheckbox_ = new QCheckBox( tr( "Show button name" ), this ) );
    showNameCheckbox_->setToolTip(
        tr( "Show this button name in taskBar.\n"
        "Note: this option overrides the\n "
        "global configuration settings" ) );

    // information layout
    mainLayout().addLayout( informationLayout_ = new QVBoxLayout(), 0 );
    informationLayout_->setMargin(0);
    informationLayout_->setSpacing(5);

    // command
    mainLayout().addLayout( vLayout = new QVBoxLayout(), 1 );
    vLayout->setMargin(0);
    vLayout->setSpacing(5);

    QFrame* frame;
    vLayout->addWidget( frame = new QFrame( this ) );
    frame->setFrameStyle( QFrame::HLine | QFrame::Sunken );

    GridLayout* gridLayout;
    vLayout->addLayout( gridLayout = new GridLayout() );
    gridLayout->setMargin(0);
    gridLayout->setSpacing(5);
    gridLayout->setMaxCount(2);

    gridLayout->addWidget( label = new QLabel( tr( "Command:" ), this ) );
    gridLayout->addWidget( commandEditor_ = new BrowsedLineEditor( this ) );
    commandEditor_->setToolTip(
        tr( "Command.\n"
        "Program executed in the background\n"
        "when clicking (left button) on the\n"
        "button.\n"
        "Add double-quotes around strings\n"
        "that contain white spaces" ) );
    label->setBuddy( commandEditor_ );

    // comments
    vLayout->addWidget( frame = new QFrame( this ) );
    frame->setFrameStyle( QFrame::HLine | QFrame::Sunken );

    vLayout->addWidget( label = new QLabel( tr( "Comments:" ), this ) );
    vLayout->addWidget( commentsEditor_ = new TextEditor( this ) );
    commentsEditor_->setToolTip(
        tr( "Button comments.\n"
        "They are displayed in the tooltip" ) );
    label->setBuddy( commentsEditor_ );

}

//________________________________________________________
TaskDialog& TaskDialog::setAttributes( const TaskAttributes& attributes )
{

    Debug::Throw( "TaskDialog::setAttributes.\n" );

    // store attributes
    attributes_ = attributes;
    if( !attributes_.creation().isValid() )
    {
        if( attributes_.modification().isValid() || attributes_.lastCalled().isValid() ) attributes_.setCreation( qMin(  attributes_.modification(), attributes_.lastCalled() ) );
        else attributes_.setCreation( TimeStamp::now() );
    }

    // update widgets
    iconEdit_->setFile( attributes_.iconName(), false );
    nameEditor_->setText( attributes_.name() );
    showNameCheckbox_->setChecked( attributes_.hasFlag( TaskAttributes::ShowName ) );
    commandEditor_->editor().setText( attributes_.command() );
    commentsEditor_->setText( attributes_.comments() );

    // add attributes information
    QFrame* frame;
    informationLayout_->addWidget( frame = new QFrame( this ) );
    frame->setFrameStyle( QFrame::HLine | QFrame::Sunken );

    GridLayout *gridLayout = new GridLayout();
    gridLayout->setMargin(0);
    gridLayout->setSpacing(5);
    gridLayout->setMaxCount(2);
    gridLayout->setColumnAlignment( 0, Qt::AlignRight|Qt::AlignVCenter );
    informationLayout_->addLayout( gridLayout );

    if( attributes_.creation().isValid() )
    {
        gridLayout->addWidget( new QLabel( "Created: ", this ) );
        gridLayout->addWidget( new QLabel( attributes_.creation().toString(), this ) );
    }

    if( attributes_.modification().isValid() )
    {
        gridLayout->addWidget( new QLabel( "Last modified: ", this ) );
        gridLayout->addWidget( new QLabel( attributes_.modification().toString(), this ) );
    }

    if( attributes_.lastCalled().isValid() )
    {
        gridLayout->addWidget( new QLabel( "Last called: ", this ) );
        gridLayout->addWidget( new QLabel( attributes_.lastCalled().toString(), this ) );
    }

    return *this;

}

//_______________________________________
const TaskAttributes& TaskDialog::attributes( void )
{

    attributes_.setName( nameEditor_->text() );
    attributes_.setCommand( commandEditor_->editor().text() );
    attributes_.setComments( commentsEditor_->toPlainText() );
    attributes_.setIconName( iconEdit_->file() );
    attributes_.setFlag( TaskAttributes::ShowName, showNameCheckbox_->isChecked() );
    return attributes_;

}
