/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "TaskListConfiguration.h"

#include "IconEngine.h"
#include "IconNames.h"
#include "InformationDialog.h"
#include "QuestionDialog.h"
#include "TaskAttributesDelegate.h"
#include "TaskDialog.h"
#include "XmlOptions.h"

#include <QHeaderView>
#include <QLayout>

//______________________________________________________
TaskListConfiguration::TaskListConfiguration( QWidget* parent, const TaskAttributes::List& taskList ):
    QWidget( parent ),
    Counter( "TaskListConfiguration" )
{
    setLayout( new QHBoxLayout() );
    layout()->setSpacing(5);
    layout()->setMargin(5);

    list_ = new TreeView( this );
    list_->setModel( &model_ );
    list_->setSortingEnabled( false );
    list_->setAllColumnsShowFocus( true );
    list_->setSelectionMode( QAbstractItemView::ContiguousSelection );

    // replace item delegate
    if( list_->itemDelegate() ) list_->itemDelegate()->deleteLater();
    list_->setItemDelegate( new TaskAttributesDelegate( this ) );

    list_->setMask( 1 );
    list_->forceAlternatingRowColors();
    list_->header()->hide();

    layout()->addWidget( list_ );

    connect( list_, SIGNAL(activated(QModelIndex)), SLOT(_edit()) );

    // vbox for buttons
    QVBoxLayout* vLayout( new QVBoxLayout() );
    vLayout->setMargin( 0 );
    vLayout->setSpacing( 5 );
    layout()->addItem( vLayout );

    // add task
    QPushButton *button;
    vLayout->addWidget( button = new QPushButton( tr( "Add Button" ), this ) );
    connect( button, SIGNAL(clicked()), this, SLOT(_add()) );
    button->setIcon( IconEngine::get( IconNames::Add ) );
    button->setToolTip( tr( "Add a new task" ) );
    addButton_ = button;

    // add separator
    vLayout->addWidget( button = new QPushButton( tr( "Add Separator" ), this ) );
    connect( button, SIGNAL(clicked()), this, SLOT(_addSeparator()) );
    button->setIcon( IconEngine::get( IconNames::Add ) );
    button->setToolTip( tr( "Add a separator" ) );
    addSeparator_ = button;

    // edit
    vLayout->addWidget( button = new QPushButton( tr( "Edit" ), this ) );
    connect( button, SIGNAL(clicked()), this, SLOT(_edit()) );
    button->setIcon( IconEngine::get( IconNames::Rename ) );
    button->setToolTip( tr( "Edit selected task" ) );
    edit_ = button;

    // remove
    vLayout->addWidget( button = new QPushButton( tr( "Remove" ), this ) );
    connect( button, SIGNAL(clicked()), this, SLOT(_remove()) );
    button->setToolTip( tr( "Remove selected task" ) );
    button->setIcon( IconEngine::get( IconNames::Remove ) );
    remove_ = button;

    vLayout->addWidget( button = new QPushButton( tr( "Move Up" ), this ) );
    connect( button, SIGNAL(clicked()), this, SLOT(_up()) );
    button->setIcon( IconEngine::get( IconNames::Up ) );
    button->setToolTip( tr( "Move up selected tasks" ) );
    moveUp_ = button;

    vLayout->addWidget( button = new QPushButton( tr( "Move Down" ), this ) );
    connect( button, SIGNAL(clicked()), this, SLOT(_down()) );
    button->setIcon( IconEngine::get( IconNames::Down ) );
    button->setToolTip( tr( "Move down selected tasks" ) );
    moveDown_ = button;

    vLayout->addStretch(1);


    // set connections
    connect( list_->selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)), SLOT(_updateButtons()) );
    connect( list_->selectionModel(), SIGNAL(currentChanged(QModelIndex,QModelIndex)), SLOT(_updateButtons()) );

    // update buttons
    _updateButtons();

    // set attributes
    setTaskList( taskList );

}

//_____________________________________________________
void TaskListConfiguration::setTaskList( const TaskAttributes::List& taskList )
{

    Debug::Throw( "TaskListConfiguration::setTaskList.\n" );
    backupTaskList_ = taskList;
    model_.set( taskList );

}

//_____________________________________________________
TaskAttributes::List TaskListConfiguration::taskAttributesList( void )
{ return model_.get(); }


//_____________________________________________________
void TaskListConfiguration::apply( void )
{ emit configurationChanged( taskAttributesList() ); }

//_____________________________________________________
void TaskListConfiguration::cancel( void )
{
    setTaskList( backupTaskAttributesList() );
    emit configurationChanged( backupTaskAttributesList() );
}

//_________________________________________________________
void TaskListConfiguration::_updateButtons( void )
{

    Debug::Throw( "TaskListConfiguration::_updateButtons.\n" );

    // get selected indexes
    TaskAttributesModel::List selection( model_.get( list_->selectionModel()->selectedRows() ) );
    bool hasSelection( !selection.empty() );
    bool hasTasks( selection.size() - std::count_if( selection.begin(), selection.end(), TaskAttributes::IsSeparatorFTor() ) > 0 );

    remove_->setEnabled( hasSelection );
    edit_->setEnabled( hasTasks );

    moveUp_->setEnabled( hasSelection && !list_->selectionModel()->isSelected( model_.index( 0, 0 ) ) );
    moveDown_->setEnabled( hasSelection && !list_->selectionModel()->isSelected( model_.index( model_.rowCount()-1, 0 ) ) );

}

//_________________________________________________________
void TaskListConfiguration::_add( void )
{

    Debug::Throw( "TaskListConfiguration::_Add.\n" );
    TaskDialog dialog( this );
    if( !dialog.exec() ) return;
    TaskAttributes attributes( dialog.attributes() );
    model_.add( attributes );

    // retrieve matching index select and make visible
    QModelIndex index( model_.index( attributes ) );
    list_->scrollTo( index );
    list_->selectionModel()->select( index, QItemSelectionModel::SelectCurrent|QItemSelectionModel::Rows );
    list_->selectionModel()->setCurrentIndex( index, QItemSelectionModel::SelectCurrent|QItemSelectionModel::Rows );

}

//_________________________________________________________
void TaskListConfiguration::_addSeparator( void )
{
    Debug::Throw( "TaskListConfiguration::_Add.\n" );
    TaskAttributes attributes;
    attributes.setFlag( TaskAttributes::IsSeparator, true );
    model_.add( attributes );

    // retrieve matching index select and make visible
    QModelIndex index( model_.index( attributes ) );
    list_->scrollTo( index );
    list_->selectionModel()->select( index, QItemSelectionModel::SelectCurrent|QItemSelectionModel::Rows );
    list_->selectionModel()->setCurrentIndex( index, QItemSelectionModel::SelectCurrent|QItemSelectionModel::Rows );

}

//_________________________________________________________
void TaskListConfiguration::_edit( void )
{

    Debug::Throw( "TaskListConfiguration::_Edit.\n" );
    TaskAttributes::List selection( model_.get( list_->selectionModel()->selectedRows() ) );
    if( selection.empty() )
    {
        InformationDialog( this, tr( "No item selected. <Edit> canceled." ) ).exec();
        return;
    }

    for( const auto& attributes:selection )
    {

        // skip if separator
        if( attributes.hasFlag( TaskAttributes::IsSeparator ) ) continue;

        TaskDialog dialog( this );
        if( !dialog.setAttributes( attributes ).exec() ) continue;

        QModelIndex index( model_.index( attributes ) );
        model_.replace( index, dialog.attributes() );

    }

    return;

}

//_________________________________________________________
void TaskListConfiguration::_remove( void )
{
    Debug::Throw( "TaskListConfiguration::_Remove.\n" );
    TaskAttributes::List selection( model_.get( list_->selectionModel()->selectedRows() ) );
    if( selection.empty() )
    {
        InformationDialog( this, tr( "No item selected. <Remove> canceled." ) ).exec();
        return;
    } else if( !QuestionDialog( this, tr( "Remove selected items ?" ) ).exec() ) return;

    // remove
    model_.remove( selection );

    return;

}

//_________________________________________________________
void TaskListConfiguration::_up( void )
{

    Debug::Throw( "TaskListConfiguration::_up.\n" );
    TaskAttributes::List selection( model_.get( list_->selectionModel()->selectedRows() ) );
    if( selection.empty() ) return;

    // retrieve selected indexes in list and store in model
    QModelIndexList selectedIndexes( list_->selectionModel()->selectedRows() );
    TaskAttributes::List selectedAttributes( model_.get( selectedIndexes ) );

    TaskAttributes::List currentAttributes( taskAttributesList() );
    TaskAttributes::List newAttributes;

    for( const auto& attributes:currentAttributes )
    {

        // check if new list is not empty, current index is selected and last index is not.
        // if yes, move.
        if(
            !( newAttributes.empty() ||
            selectedIndexes.indexOf( model_.index( attributes ) ) == -1 ||
            selectedIndexes.indexOf( model_.index( newAttributes.back() ) ) != -1
            ) )
        {
            TaskAttributes last( newAttributes.back() );
            newAttributes.removeLast();
            newAttributes.append( attributes );
            newAttributes.append( last );
        } else newAttributes.append( attributes );

    }

    model_.set( newAttributes );

    // restore selection
    list_->selectionModel()->clear();
    for( const auto& attributes:selectedAttributes )
    { list_->selectionModel()->select( model_.index( attributes ), QItemSelectionModel::Select|QItemSelectionModel::Rows ); }

    return;

}

//_________________________________________________________
void TaskListConfiguration::_down( void )
{

    Debug::Throw( "TaskListConfiguration::_down.\n" );
    TaskAttributes::List selection( model_.get( list_->selectionModel()->selectedRows() ) );
    if( selection.empty() ) return;

    // retrieve selected indexes in list and store in model
    QModelIndexList selectedIndexes( list_->selectionModel()->selectedIndexes() );
    TaskAttributes::List selectedAttributes( model_.get( selectedIndexes ) );

    TaskAttributes::List currentAttributes( TaskListConfiguration::taskAttributesList() );
    TaskAttributes::List newAttributes;
    TaskAttributes::ListIterator iter( currentAttributes );
    iter.toBack();
    while( iter.hasPrevious() )
    {
        const TaskAttributes& current( iter.previous() );
        if(
            !( newAttributes.empty() ||
            selectedIndexes.indexOf( model_.index( current ) ) == -1 ||
            selectedIndexes.indexOf( model_.index( newAttributes.front() ) ) != -1
            ) )
        {

            TaskAttributes first( newAttributes.front() );
            newAttributes.removeFirst();
            newAttributes.prepend( current );
            newAttributes.prepend( first );

        } else newAttributes.prepend( current );
    }

    model_.set( newAttributes );

    // restore selection
    list_->selectionModel()->clear();
    for( const auto& attributes:selectedAttributes )
    { list_->selectionModel()->select( model_.index( attributes ), QItemSelectionModel::Select|QItemSelectionModel::Rows ); }

    return;

}
